#pylint: disable=logging-fstring-interpolation
"""
Visualization of a capability region.
"""

import logging
import gc
from collections import deque

import tkinter as tk
from tkinter import ttk

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
from matplotlib.patches import Polygon

from hynet.scenario.capability import CapRegion, HalfSpace
from hynet.visual.capability.utilities import (Axis,
                                               Bound,
                                               LinearFunction,
                                               Orientation,
                                               Point)
from hynet.visual.capability.settings import SettingsView

_log = logging.getLogger(__name__)

MESSAGE_DISAPPEAR_WAIT_TIME = 2500

SAVE_FILETYPES = (
    ("png files", "*.png"), ("pdf files", "*.pdf"), ("ps files", "*.ps"),
    ("eps files", "*.eps"), ("svg files", "*.svg"), ("all files", "*.*")
)


class Window(ttk.Frame):  # pylint: disable=too-many-ancestors
    """This class creates a window that visualizes a capability region."""
    SCALING_FACTOR = 1
    CAPABILITY_REGION_COLOR = '#BBBBBB'
    HALFSPACES_COLOR = '#777777'
    POLYGON_EDGE_COLOR = '#777777'
    POLYGON_FILL_COLOR = '#BAEBAC'

    # this map stores the matplotlib lines as Axis x Bound -> line associations
    _limit_lines = {}
    # this map stores the matplotlib lines as
    # {LEFT, RIGHT} x {TOP, BOTTOM} -> line associations
    _halfspace_lines = {}
    _polygon = []
    _callback_queue = deque()

    def __init__(self, master=None, capability_region=CapRegion(), edit=True):
        super().__init__(master, padding='3 3 12 12')
        self._master = master
        self._cap_region = capability_region
        self.grid(row=0, column=0, sticky=tk.N + tk.S + tk.E + tk.W)
        self.fig = Figure(tight_layout=True)
        self._ax = self.fig.add_subplot(1, 1, 1)
        self._configure_canvas()
        self._canvas = FigureCanvasTkAgg(self.fig, master=master)
        self._canvas.show()
        self._canvas.get_tk_widget().grid(row=0, column=0, padx=0, pady=0,
                                          sticky=[tk.N, tk.S, tk.E, tk.W])
        self._limit_lines = {}
        self._halfspace_lines = {}
        self._polygon = []
        self._callback_queue = deque()

        if edit:
            self._settings_view = SettingsView(master=master,
                                               cap_region=capability_region)
            self._settings_view.container.grid(row=0, column=1,
                                               padx=(20, 0), pady=0,
                                               sticky=[tk.W, tk.E, tk.S, tk.N])
            self._setup_capabilityregion_callbacks()
            self._setup_halfspace_callbacks()
            self._settings_view.power_factor.set_callback(
                self._update_powerfactor)
            self._settings_view.error_callback = self._show_message

            self._message_label = tk.Label(master=self._master, padx=0, pady=0,
                                           anchor=tk.S, height=2, fg="red")
            self._message_label.grid(row=1, column=0, padx=10, pady=(0, 10),
                                     sticky=[tk.N, tk.S, tk.E, tk.W])

        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.rowconfigure(0, weight=1)
        self._update_view()

    def _update_powerfactor(self, value):
        try:
            self._cap_region.add_power_factor_limit(value)
            # update the values from the capability region
            self._settings_view.update_from_capability_region()
        except ValueError as error:
            self._show_message(str(error))
        self._update_view()

    def _show_message(self, message, undo_callback=None):
        def _callback_now():
            self._message_label.config(text=message)
            _log.warning(message)

            _log.debug(
                f"_callback_now: Length of queue: {len(self._callback_queue)}")
            if self._callback_queue:
                next_callback = self._callback_queue.popleft()
                self.after(MESSAGE_DISAPPEAR_WAIT_TIME, next_callback)

        def _delayed_callback():
            _log.debug(
                f"Delayed_call: Length of queue: {len(self._callback_queue)} "
                f"With callback?: {not (undo_callback is None)}")
            if undo_callback:
                undo_callback()
            self._message_label.config(text='')
            if self._callback_queue:
                next_callback = self._callback_queue.popleft()
                self.after(MESSAGE_DISAPPEAR_WAIT_TIME, next_callback)

        _log.debug(f"Length of queue: {len(self._callback_queue)}")
        if self._callback_queue:
            self._callback_queue.append(_callback_now)
            self._callback_queue.append(_delayed_callback)
        else:
            # the order is important!
            self._callback_queue.append(_delayed_callback)
            _callback_now()

    def _update_view(self):
        self._update_polygon()
        self._draw_capability_region()
        # self.draw_halfspaces()
        self._ax.relim(visible_only=True)
        self._ax.autoscale()
        self._canvas.draw()

    def _configure_canvas(self):
        self._ax.autoscale()
        self._ax.grid(True, which='both')

        self._ax.set_xlabel('P', horizontalalignment="center", x=1.0)
        y_label = self._ax.set_ylabel('Q', horizontalalignment="center", y=1.0)
        y_label.set_rotation(0)
        # turn off the right spine/ticks
        self._ax.spines['right'].set_color('none')
        # turn off the top spine/ticks
        self._ax.spines['top'].set_color('none')

        # draw axes at (0,0)
        self._ax.spines['left'].set_position(('data', 0.0))
        self._ax.spines['bottom'].set_position(('data', 0.0))

        # equally distant axes
        self._ax.axis('equal')

    def _setup_capabilityregion_callbacks(self):
        # arrow heads: http://felix11h.github.io/blog/arrowheads-matplotlib
        def _create_value_change_callback(capability_region, axis, bound):
            def _callback(value):  # pylint: disable=too-many-branches
                axis_tuple = (axis, bound)
                if axis_tuple == (Axis.REACTIVE_POWER, Bound.MIN):
                    if capability_region.q_max < value:
                        raise ValueError(
                            f"Please choose a smaller value.\n{value} "
                            f"exceeds the upper bound on reactive power.")
                    else:
                        capability_region.q_min = value
                elif axis_tuple == (Axis.REACTIVE_POWER, Bound.MAX):
                    if capability_region.q_min > value:
                        raise ValueError(
                            f"Please choose a larger value.\n{value} "
                            f"is below the lower bound on reactive power.")
                    else:
                        capability_region.q_max = value
                elif axis_tuple == (Axis.ACTIVE_POWER, Bound.MIN):
                    if capability_region.p_max < value:
                        raise ValueError(
                            f"Please choose a smaller value.\n{value} "
                            f"exceeds the upper bound on active power.")
                    else:
                        capability_region.p_min = value
                elif axis_tuple == (Axis.ACTIVE_POWER, Bound.MAX):
                    if capability_region.q_min > value:
                        raise ValueError(
                            f"Please choose a larger value.\n{value} "
                            f"is below the lower bound on active power.")
                    else:
                        capability_region.p_max = value
                else:
                    ValueError("No implementation for axis " + str(axis) +
                               " and bound " + str(bound))
                p_same = self._cap_region.p_min == self._cap_region.p_max
                q_same = self._cap_region.q_min == self._cap_region.q_max
                if p_same or q_same:
                    self._settings_view.deactivate_all_halfspaces()
                    capability_region.rt = None
                    capability_region.lt = None
                    capability_region.rb = None
                    capability_region.lb = None
                self._update_view()

            return _callback

        self._settings_view.active_power.left_entry.set_callback(
            _create_value_change_callback(
                capability_region=self._cap_region,
                axis=Axis.ACTIVE_POWER,
                bound=Bound.MIN))
        self._settings_view.active_power.right_entry.set_callback(
            _create_value_change_callback(
                capability_region=self._cap_region,
                axis=Axis.ACTIVE_POWER,
                bound=Bound.MAX))
        self._settings_view.reactive_power.left_entry.set_callback(
            _create_value_change_callback(
                capability_region=self._cap_region,
                axis=Axis.REACTIVE_POWER,
                bound=Bound.MIN))
        self._settings_view.reactive_power.right_entry.set_callback(
            _create_value_change_callback(
                capability_region=self._cap_region,
                axis=Axis.REACTIVE_POWER,
                bound=Bound.MAX))
        self._settings_view.setup_halfspace_callbacks(self._update_view)

    def _draw_capability_region(self):

        def _parallel_to_axis(parallel_to, ends, orientation):
            if not len(ends) == 2:
                raise ValueError("A line must have 2 ends.")
            if orientation is Axis.REACTIVE_POWER:
                return ([Window.SCALING_FACTOR * ends[0],
                         Window.SCALING_FACTOR * ends[1]],
                        [parallel_to, parallel_to])
            if orientation is Axis.ACTIVE_POWER:
                return ([parallel_to, parallel_to],
                        [Window.SCALING_FACTOR * ends[0],
                         Window.SCALING_FACTOR * ends[1]])
            raise ValueError("Unknown orientation: " + orientation)

        def _draw_region_part(capability_region, axis, bound):
            """
            Parameters
            ----------
            capability_region: CapRegion
                The capability region specification.
            axis: Axis
                Either Axis.POWER or Axis.REACTIVE_POWER for the active
                and reactive power, respectively.
            bound: Bound
                Bound.MIN or Bound.MAX to specify the bound.

            Returns
            -------
            The line that was added to the plot.
            """
            limits = []
            bound_value = None
            if axis is Axis.ACTIVE_POWER:
                limits = [self._cap_region.q_min,
                          self._cap_region.q_max]
                if bound is Bound.MAX:
                    bound_value = capability_region.p_max
                elif bound is Bound.MIN:
                    bound_value = capability_region.p_min
                else:
                    raise ValueError("Unsupported Bound: " + bound)
            elif axis is Axis.REACTIVE_POWER:
                limits = [self._cap_region.p_min,
                          self._cap_region.p_max]
                if bound is Bound.MAX:
                    bound_value = capability_region.q_max
                elif bound is Bound.MIN:
                    bound_value = capability_region.q_min
                else:
                    raise ValueError("Unsupported Bound: " + bound)
            points = _parallel_to_axis(bound_value, limits, axis)

            return self._ax.plot(*points,
                                 color=Window.CAPABILITY_REGION_COLOR,
                                 zorder=0)[0]

        # remove all limit lines / points
        for axis, bound in [(a, b) for a in Axis for b in Bound]:
            if (axis, bound) in self._limit_lines:
                line = self._limit_lines[(axis, bound)]
                line.remove()
                del line
                del self._limit_lines[(axis, bound)]
        if (self._cap_region.p_min == self._cap_region.p_max) and \
            (self._cap_region.q_min == self._cap_region.q_max):
            point = self._ax.plot([self._cap_region.p_min],
                                  [self._cap_region.q_min], marker='o',
                                  linestyle='--', color='r')[0]
            self._limit_lines[(Axis.ACTIVE_POWER, Bound.MIN)] = point
        elif self._cap_region.p_min == self._cap_region.p_max:
            # draw only the line of p
            line = _draw_region_part(axis=Axis.ACTIVE_POWER, bound=Bound.MIN,
                                     capability_region=self._cap_region)
            self._limit_lines[(Axis.REACTIVE_POWER, Bound.MIN)] = line
        elif self._cap_region.q_min == self._cap_region.q_max:
            line = _draw_region_part(axis=Axis.REACTIVE_POWER, bound=Bound.MIN,
                                     capability_region=self._cap_region)
            self._limit_lines[(Axis.REACTIVE_POWER, Bound.MIN)] = line
        else:
            # create the rectangular capability region lines
            for axis, bound in [(a, b) for a in Axis for b in Bound]:
                line = _draw_region_part(axis=axis, bound=bound,
                                         capability_region=self._cap_region)
                self._limit_lines[(axis, bound)] = line

    def _draw_halfspaces(self):
        halfspace_options = {
            (Orientation.RIGHT, Orientation.TOP): self._cap_region.rt,
            (Orientation.LEFT, Orientation.TOP): self._cap_region.lt,
            (
                Orientation.RIGHT,
                Orientation.BOTTOM): self._cap_region.rb,
            (Orientation.LEFT, Orientation.BOTTOM): self._cap_region.lb}
        for (p, q), halfspace in halfspace_options.items():
            line_tuple = (p, q)
            line = self._halfspace_lines.get(line_tuple)
            if not halfspace and line is None:
                continue

            # now we remove any lines we need to remove as halfspace is None
            # but a line for this halfspace still exists
            # but during initialising the line is still None
            if line:
                del self._halfspace_lines[line_tuple]
                line.remove()
                del line

            # and add a line back, if the halfspace exists
            if halfspace:
                points = self._halfspace_points(p_orientation=p,
                                                q_orientation=q)
                line = self._ax.plot(*points,
                                     color=Window.HALFSPACES_COLOR,
                                     zorder=0)[0]
                self._halfspace_lines[line_tuple] = line

    def _halfspace_points(self, p_orientation, q_orientation):
        #   l/t     ^ q_max   r/t
        #           |
        # p_min     |         p_max
        # p --------|-------->
        #           |
        #           |
        #   l/b     q q_min   r/b
        if q_orientation is Orientation.TOP:
            q = self._cap_region.q_max
            possible_halfspaces = (
                self._cap_region.lt, self._cap_region.rt)
        elif q_orientation is Orientation.BOTTOM:
            q = self._cap_region.q_min
            possible_halfspaces = (
                self._cap_region.lb, self._cap_region.rb)
        else:
            raise ValueError(str(q_orientation) +
                             " is not allowed for q_orientation.")

        if p_orientation is Orientation.RIGHT:
            p = self._cap_region.p_max
            halfspace = possible_halfspaces[1]
        elif p_orientation is Orientation.LEFT:
            p = self._cap_region.p_min
            halfspace = possible_halfspaces[0]
        else:
            raise ValueError(str(p_orientation) +
                             " is not allowed for p_orientation.")
        if halfspace is None:
            return None

        # now we can be sure that the halfspace exists and we have
        # the right values
        limits = Point(p, q)
        line_functions = LinearFunction(halfspace, limits)

        x_points = [limits.p * Window.SCALING_FACTOR, limits.p,
                    line_functions.f_1(limits.q),
                    line_functions.f_1(limits.q * Window.SCALING_FACTOR)]
        y_points = [line_functions.f(limits.p * Window.SCALING_FACTOR),
                    line_functions.f(limits.p), limits.q,
                    limits.q * Window.SCALING_FACTOR]
        return x_points, y_points

    def _setup_halfspace_callbacks(self):
        halfspace_options = {(Orientation.RIGHT, Orientation.TOP):
                                 self._settings_view.rt_half_space,
                             (Orientation.LEFT, Orientation.TOP):
                                 self._settings_view.lt_half_space,
                             (Orientation.RIGHT, Orientation.BOTTOM):
                                 self._settings_view.rb_half_space,
                             (Orientation.LEFT, Orientation.BOTTOM):
                                 self._settings_view.lb_half_space}

        def _create_callback(p_orientation, q_orientation, slope):
            def _callback(value):
                # update the value on the capability region
                def _get_new_halfspace(old):
                    if old:
                        if slope:
                            return HalfSpace(slope=value, offset=old.offset)
                        return HalfSpace(slope=old.slope, offset=value)
                    return None

                line_tuple = (p_orientation, q_orientation)
                try:
                    if line_tuple == (Orientation.RIGHT, Orientation.TOP):
                        self._cap_region.rt = _get_new_halfspace(
                            self._cap_region.rt)
                    elif line_tuple == (Orientation.RIGHT, Orientation.BOTTOM):
                        self._cap_region.rb = _get_new_halfspace(
                            self._cap_region.rb)
                    elif line_tuple == (Orientation.LEFT, Orientation.TOP):
                        self._cap_region.lt = _get_new_halfspace(
                            self._cap_region.lt)
                    elif line_tuple == (Orientation.LEFT, Orientation.BOTTOM):
                        self._cap_region.lb = _get_new_halfspace(
                            self._cap_region.lb)
                    else:
                        raise ValueError(
                            str(line_tuple) + " is not a valid configuration!")
                except ValueError as error:
                    half_plane = halfspace_options[line_tuple]
                    halfspace = None
                    if line_tuple == (Orientation.RIGHT, Orientation.TOP):
                        halfspace = self._cap_region.rt
                    elif line_tuple == (Orientation.RIGHT, Orientation.BOTTOM):
                        halfspace = self._cap_region.rb
                    elif line_tuple == (Orientation.LEFT, Orientation.TOP):
                        halfspace = self._cap_region.lt
                    elif line_tuple == (Orientation.LEFT, Orientation.BOTTOM):
                        halfspace = self._cap_region.lb

                    if slope:
                        undo = half_plane.right_entry.error_state(
                            old_value=halfspace.slope)
                    else:
                        undo = half_plane.left_entry.error_state(
                            old_value=halfspace.offset)
                    self._show_message(str(error), undo)
                self._update_view()

            return _callback

        for (p, q), descriptive_entry in halfspace_options.items():
            left_callback = _create_callback(p_orientation=p, q_orientation=q,
                                             slope=False)
            right_callback = _create_callback(p_orientation=p, q_orientation=q,
                                              slope=True)

            descriptive_entry.left_entry.set_callback(left_callback)
            descriptive_entry.right_entry.set_callback(right_callback)

    def _update_polygon(self):
        try:
            polygon_vertices = self.compute_polygon_vertices(self._cap_region)
        except ValueError as error:
            self._show_message(str(error))
        # update the polygon
        if self._polygon:
            for polygon in self._polygon:
                polygon.remove()
            self._polygon = []

        if polygon_vertices:
            polygon = Polygon(np.array(polygon_vertices),
                              closed=True,
                              facecolor=Window.POLYGON_FILL_COLOR,
                              zorder=0)
            self._ax.add_patch(polygon)
            self._polygon.append(polygon)
            polygon_outline = Polygon(np.array(polygon_vertices),
                                      closed=True,
                                      edgecolor=Window.POLYGON_EDGE_COLOR,
                                      facecolor='none',
                                      zorder=4)
            self._ax.add_patch(polygon_outline)
            self._polygon.append(polygon_outline)

    def display_operating_point(self, point):
        """
        Display an operating point in the P/Q-plane of the capability region.

        Parameters
        ----------
        point: tuple(p,q)
            This tuple represents the operating point as a (p,q)-tuple of
            floats.
        """
        self._ax.plot([point[0]], [point[1]], label='Operating Point',
                      marker='+', linestyle='--', color='r')
        self._update_view()

    @staticmethod
    def compute_polygon_vertices(cap_region):
        """
        Compute the vertices of the capability region polygon.

        This function computes the intersections of the four lines that delimit
        the capability region. The algorithm starts with the left-top half
        space and continues clock-wise with the other half-spaces.

        The algorithm assumes that ``p_max != p_min`` and ``q_max != q_min``.
        Furthermore, the assumptions on the slopes and offsets must be
        satisfied as well. If the preconditions are met, the function returns
        a list of all vertices that specify the polygon, in a clock-wise
        ordering.

        Parameters
        ----------
        cap_region: CapRegion
            The capability region that shall be drawn.

        Returns
        -------
        list[tuple(p,q)]
            List of polygon vertices.
        """
        # pylint: disable=too-many-locals,too-many-statements,too-many-branches

        # l/t     ^ q_max   r/t
        #           |
        # p_min     |         p_max
        # p --------|-------->
        #           |
        #           |
        #   l/b     q q_min   r/b
        # start with lt and compute the q value with p_min
        # then continue with the intersection / the top points of lt and rt.
        # followed by the p_max point of rt, then do the same for the bottom
        # part
        # Remember that the halfspaces could be empty!
        # The offset is required to be positive!
        all_top_halfspaces_present = cap_region.lt is not None and \
                                     cap_region.rt is not None
        all_bottom_halfspaces_present = cap_region.lb is not None and \
                                        cap_region.rb is not None
        polygon_vertices = []
        limits_lt = Point(cap_region.p_min, cap_region.q_max)
        limits_rt = Point(cap_region.p_max,
                          cap_region.q_max)
        limits_rb = Point(cap_region.p_max,
                          cap_region.q_min)
        limits_lb = Point(cap_region.p_min,
                          cap_region.q_min)

        def _point_in_bounds(point):
            q_in_bounds = cap_region.q_min <= point.q <= cap_region.q_max
            p_in_bounds = cap_region.p_min <= point.p <= cap_region.p_max
            return q_in_bounds and p_in_bounds

        def _compute_and_limit(linear_function, limits, compute_coordinate):
            if compute_coordinate is Axis.ACTIVE_POWER:
                point = Point(linear_function.f_1(limits.q), limits.q)
                if not _point_in_bounds(point):
                    point = Point(limits.p, linear_function.f(limits.p))
            elif compute_coordinate is Axis.REACTIVE_POWER:
                point = Point(limits.p, linear_function.f(limits.p))
                if not _point_in_bounds(point):
                    point = Point(linear_function.f_1(limits.q), limits.q)
            else:
                raise ValueError(f"{compute_coordinate} is not implemented!")
            if not _point_in_bounds(point):
                _log.warning(f"Point {point} for halfspace with offset "
                             f"{linear_function.offset} and "
                             f"{linear_function.slope} slope is not in "
                             f"the bounds of the capability region. "
                             f"We will limit it to the limits of the "
                             f"capability region.")
                return limits
            return point

        if all_top_halfspaces_present:
            plf_lt = LinearFunction(cap_region.lt, limits_lt)
            plf_rt = LinearFunction(cap_region.rt, limits_rt)

            polygon_vertices.append(
                _compute_and_limit(plf_lt, limits_lt, Axis.REACTIVE_POWER))
            intersection = plf_lt.intersect_with(plf_rt)
            if intersection:
                # should be the case in all cases were the halfspace is present
                if _point_in_bounds(intersection):
                    polygon_vertices.append(intersection)
                else:
                    # compute the capped intersection with q_max (as the
                    # intersection of the two halfspace function is out of bounds
                    lt_top_point = _compute_and_limit(plf_lt, limits_lt,
                                                      Axis.ACTIVE_POWER)
                    rt_top_point = _compute_and_limit(plf_rt, limits_rt,
                                                      Axis.ACTIVE_POWER)
                    polygon_vertices.append(lt_top_point)
                    polygon_vertices.append(rt_top_point)
            else:
                message = "The top halfspaces are parallel, " \
                          "please check your input values!"
                raise ValueError(message)
            polygon_vertices.append(
                _compute_and_limit(plf_rt, limits_rt, Axis.REACTIVE_POWER))
        else:
            if cap_region.lt:
                plf = LinearFunction(cap_region.lt, limits_lt)
                lt_top_point = _compute_and_limit(plf, limits_lt,
                                                  Axis.REACTIVE_POWER)
                lt_bottom_point = _compute_and_limit(plf, limits_lt,
                                                     Axis.ACTIVE_POWER)
                rt_top_point = _compute_and_limit(plf, limits_rt,
                                                  Axis.REACTIVE_POWER)

                polygon_vertices.append(lt_top_point)
                polygon_vertices.append(lt_bottom_point)
                # append replacement point for rt
                if rt_top_point == lt_bottom_point:
                    polygon_vertices.append(limits_rt)
                else:
                    polygon_vertices.append(rt_top_point)

            elif cap_region.rt:
                plf = LinearFunction(cap_region.rt, limits_rt)
                # append replacement point for lt
                lt_top_point = _compute_and_limit(plf, limits_lt,
                                                  Axis.REACTIVE_POWER)
                rt_top_point = _compute_and_limit(plf, limits_rt,
                                                  Axis.ACTIVE_POWER)
                rt_bottom_point = _compute_and_limit(plf, limits_rt,
                                                     Axis.REACTIVE_POWER)
                if lt_top_point == rt_top_point:
                    polygon_vertices.append(limits_lt)
                else:
                    polygon_vertices.append(lt_top_point)

                polygon_vertices.append(rt_top_point)
                polygon_vertices.append(rt_bottom_point)
            else:
                polygon_vertices.append(limits_lt)
                polygon_vertices.append(limits_rt)

        if all_bottom_halfspaces_present:
            # p_min     |         p_max
            # p --------|-------->
            #           |
            #           |
            #   l/b     q q_min   r/b
            plf_rb = LinearFunction(cap_region.rb, limits_rb)
            plf_lb = LinearFunction(cap_region.lb, limits_lb)

            polygon_vertices.append(
                _compute_and_limit(plf_rb, limits_rb, Axis.REACTIVE_POWER))

            # compute the intersection of the two functions and determine if it
            # is still in bounds (i.e., the q is smaller than q_min
            intersection = plf_rb.intersect_with(plf_lb)
            if intersection:
                if _point_in_bounds(intersection):
                    polygon_vertices.append(intersection)
                else:
                    # first intersect rb with q_min and then lb
                    point_limited_rb_intersection = \
                        _compute_and_limit(plf_rb, limits_rb,
                                           Axis.ACTIVE_POWER)
                    polygon_vertices.append(point_limited_rb_intersection)
                    point_limited_lb_intersection = \
                        _compute_and_limit(plf_lb, limits_lb,
                                           Axis.ACTIVE_POWER)
                    polygon_vertices.append(point_limited_lb_intersection)
            else:
                message = "The bottom halfspaces are parallel, " \
                          "please check your input values!"
                raise ValueError(message)
            lb_end = _compute_and_limit(plf_lb, limits_lb, Axis.REACTIVE_POWER)
            polygon_vertices.append(lb_end)
        else:
            if cap_region.rb:
                plf_rb = LinearFunction(cap_region.rb, limits_rb)
                point_rb_top = _compute_and_limit(plf_rb, limits_rb,
                                                  Axis.REACTIVE_POWER)
                point_rb_bottom = _compute_and_limit(plf_rb, limits_rb,
                                                     Axis.ACTIVE_POWER)
                # compute replacement lb points
                point_lb_bottom = _compute_and_limit(plf_rb, limits_lb,
                                                     Axis.REACTIVE_POWER)
                polygon_vertices.append(point_rb_top)
                polygon_vertices.append(point_rb_bottom)

                if point_rb_bottom == point_lb_bottom:
                    polygon_vertices.append(limits_lb)
                else:
                    polygon_vertices.append(point_lb_bottom)

            elif cap_region.lb:
                plf_lb = LinearFunction(cap_region.lb, limits_lb)
                # limit value to bounds
                # compute replacement rb points
                point_rb_bottom = _compute_and_limit(plf_lb, limits_rb,
                                                     Axis.REACTIVE_POWER)
                point_lb_bottom = _compute_and_limit(plf_lb, limits_lb,
                                                     Axis.ACTIVE_POWER)
                point_lb_top = _compute_and_limit(plf_lb, limits_lb,
                                                  Axis.REACTIVE_POWER)

                if point_rb_bottom == point_lb_bottom:
                    polygon_vertices.append(limits_rb)
                else:
                    polygon_vertices.append(point_rb_bottom)
                polygon_vertices.append(point_lb_bottom)
                polygon_vertices.append(point_lb_top)
            else:
                polygon_vertices.append(limits_rb)
                polygon_vertices.append(limits_lb)

        return polygon_vertices

    @staticmethod
    def show(capability_region, edit=True, operating_point=None):
        """
        Create and show a capability region visualizer window.

        Parameters
        ----------
        capability_region: CapRegion
        edit: bool, optional
            If True (default), editing of the specified capability region is
            enabled.
        operating_point: tuple(p,q), optional
            If provided, this operating point is shown by a marker in the
            P/Q-plane.
        """
        # If we do not close the open figures, the settings view does not work
        # properly, strangely...maybe we try to find another fix sometime :)
        plt.close('all')

        root = tk.Tk()
        root.title("Capability Region Visualizer - Press [ESC] to exit.")
        tk.Grid.rowconfigure(root, 0, weight=1)
        tk.Grid.columnconfigure(root, 0, weight=1)
        app = Window(root, capability_region, edit)
        if operating_point:
            app.display_operating_point(operating_point)

        # Bind the escape key to exit the application
        def _quit(event):  # pylint: disable=unused-argument
            root.quit()     # Stop the main loop
            root.destroy()  # Close the window
        root.bind("<Escape>", _quit)

        # Set up the "Export as image" button
        button_frame = tk.Frame(root, padx=0, pady=0, bg="white")

        def _save_figure():
            from pathlib import Path
            from tkinter import filedialog
            file_path = filedialog.asksaveasfilename(initialdir=Path.home(),
                                                     filetypes=SAVE_FILETYPES,
                                                     title="Export as image...")
            if file_path:
                app.fig.savefig(file_path)

        save_fig = tk.Button(button_frame,
                             text='Export as image...',
                             command=_save_figure,
                             padx=10,
                             bg="white")
        save_fig.grid(row=0, column=0, sticky=[tk.N, tk.S, tk.E, tk.W])

        if edit:
            button_frame.grid(row=1, column=1, padx=(20, 0), pady=(15, 10),
                              sticky=[tk.N, tk.S, tk.E, tk.W])
            root.update()
            root.minsize(width=750, height=610)
        else:
            button_frame.grid(row=1, column=0, padx=10, pady=(0, 10),
                              sticky=tk.E)
            root.update()
            root.minsize(width=500, height=350)

        while True:
            try:
                root.mainloop()
                break
            except UnicodeDecodeError:
                # See https://stackoverflow.com/questions/16995969/
                # inertial-scrolling-in-mac-os-x-with-tkinter-and-python
                pass

        # REMARK: After closing the window, we received runtime errors that
        # stated "RuntimeError: main thread is not in main loop". This error
        # appears when tk is not run in the main thread, see
        # https://stackoverflow.com/questions/14694408/runtimeerror-main-thread-is-not-in-main-loop
        # However, GUI was run in the main thread, but the garbage collection
        # of the window by one of the threads spawn during the OPF computation,
        # see also
        # https://stackoverflow.com/questions/44781806/tkinter-objects-being-garbage-collected-from-the-wrong-thread
        # To avoid this issue, we enforce the garbage collection in the main
        # thread here:
        del root
        gc.collect()



#pylint: disable=bad-whitespace
"""
Support for the generation of scenarios of exemplary days.
"""

import logging

import numpy as np

from hynet.data.interface import initialize_database, save_scenario

_log = logging.getLogger(__name__)

LOAD_WEIGHTS = {
    'winter_weekday':
        np.array([ 67,  63, 60, 59,  59,  60, 74,  86, 95, 96, 96,  95, 95,
                   95,  93, 94, 99, 100, 100, 96,  91, 83, 73, 63]) / 100,
    'winter_weekend':
        np.array([ 78,  72, 68, 66,  64,  65, 66,  70, 80, 88, 90,  91, 90,
                   88,  87, 87, 91, 100,  99, 97,  94, 92, 87, 81]) / 100,
    'summer_weekday':
        np.array([ 64,  60, 58, 56,  56,  58, 64,  76, 87, 95, 99, 100, 99,
                  100, 100, 97, 96,  96,  93, 92,  92, 93, 87, 72]) / 100,
    'summer_weekend':
        np.array([ 74,  70, 66, 65,  64,  62, 62,  66, 81, 86, 91,  93, 93,
                   92,  91, 91, 92,  94,  95, 95, 100, 93, 88, 80]) / 100
}


def initialize_database_with_example_days(database, base_case):
    """
    Initialize the database with the base case and add exemplary day scenarios.

    The added day scenarios are exemplary winter and summer weekdays/weekends
    obtained by scaling the base case load according to the data provided in
    [1]_, Table 4.

    Parameters
    ----------
    database : DBConnection
        Connection to the destination database. This database must be empty.
    base_case : Scenario
        Base case scenario for the new database.

    Raises
    ------
    ValueError
        In case that the destination database is not empty or the provided
        scenario data exhibits integrity or validity issues.

    References
    ----------
    .. [1]  C. Grigg et al., "The IEEE Reliability Test System - 1996. A report
            prepared by the Reliability Test System Task Force of the
            Application of Probability Methods Subcommittee," in IEEE Trans.
            Power Systems, vol. 14, no. 3, pp. 1010-1020, Aug. 1999.
    """
    base_case.verify()
    if not database.empty:
        raise ValueError("The destination database is not empty.")
    initialize_database(database, base_case)
    add_example_days(database, base_case)


def add_example_days(database, base_case):
    """Add the exemplary day scenarios to the database."""
    for scenario_name, weights in LOAD_WEIGHTS.items():
        scenario_name = scenario_name.replace('_', ' ').title()
        for time, weight in enumerate(weights):
            scenario_new = base_case.copy()
            scenario_new.name = scenario_name
            scenario_new.time = time
            scenario_new.bus['load'] *= weight
            save_scenario(database, scenario_new)



"""
Graph-related utilities.
"""

import logging

import numpy as np
import scipy.sparse
import scipy.sparse.csgraph

from hynet.types_ import hynet_int_, hynet_float_
from hynet.utilities.base import create_sparse_matrix

_log = logging.getLogger(__name__)


def eliminate_parallel_edges(edges):
    """
    Eliminates parallel edges from the graph and returns those edges.

    The edges are considered undirected, i.e., and the source and destination
    node are interchangeable. In the returned edges, the source is set to the
    adjacent node with the lower node number.
    """
    E = np.column_stack((edges[0], edges[1]))
    if E.shape[0] != 0:
        E.sort(axis=1)
        E = np.unique(E, axis=0)
    return E[:, 0], E[:, 1]


def traverse_graph(nodes, edges, callback, roots=None, auto_root=True):
    """
    Traverse the graph and run a callback at every node.

    The graph is traversed in a depth-first fashion and, at every visited node,
    the provided callback function is called, which must exhibit the signature::

        def callback(node, node_pre, cycle)

    Therein, ``node`` is the currently visited node, ``node_pre`` is the
    previously visited node (or ``numpy.nan`` if the current node is the
    component's root, i.e., the traversal entered a new component), and
    ``cycle`` is ``True`` if the current node was already visited, i.e., the
    edge traversed from the previous to the current node closes a cycle. The
    callback function returns an abort flag, i.e., if the callback returns
    ``True``, the traversal is aborted.

    **Remark:** This function is most efficient if the nodes are numbered
    consecutively from zero to the number of nodes minus one. Negative node
    numbers are not supported.

    Parameters
    ----------
    nodes : numpy.ndarray[.hynet_int_]
        Node numbers of the graph.
    edges : (numpy.ndarray[.hynet_int_], numpy.ndarray[.hynet_int_])
        Edges of the graph in terms of node number tuples.
    callback : function
        Node visit callback function.
    roots : numpy.ndarray[.hynet_int_], optional
        Numbers of the root nodes of the graph.
    auto_root : bool, optional
        If True (default), components without a root node are assigned an
        arbitrary node as its root.
    """
    # timer = Timer()
    # debug_message = ("Graph traversal with '"
    #                  + callback.__name__ + "' ({:.3f} sec.)")
    roots = np.ndarray(0) if roots is None else roots

    if not nodes.size:
        return

    (e_src, e_dst) = eliminate_parallel_edges(edges)

    if min(nodes) < 0:
        raise RuntimeError('Negative node numbers are not supported.')

    if not (np.all(np.isin(e_src, nodes)) and np.all(np.isin(e_dst, nodes))):
        raise RuntimeError('The edges are referencing non-existent nodes.')

    if not np.all(np.isin(roots, nodes)):
        raise RuntimeError('The roots are referencing non-existent nodes.')

    # REMARK: As this implementation is robust against duplicate entries in
    # 'nodes', this data validity check is omitted.

    N = max(nodes) + 1
    stack = np.zeros(N, dtype=hynet_int_)  # Stack of nodes to be visited
    visited = np.zeros(N, dtype=bool)      # Indicates if a node was visited
    root_idx = 0                           # Index of the next root to process

    # Set gaps in the node numbering to 'visited'
    visited[np.setdiff1d(range(N), nodes)] = True

    while not np.all(visited):
        # Ignore ambiguous root nodes
        while root_idx < len(roots) and visited[roots[root_idx]]:
            _log.warning("Ambiguous root {} is ignored.".format(roots[root_idx]))
            root_idx += 1

        # Set root node for next component to traverse
        stack_top = 0
        if root_idx < len(roots):
            stack[stack_top] = roots[root_idx]
            root_idx += 1
        elif auto_root:
            stack[stack_top] = np.where(~visited)[0][0]
        else:
            raise RuntimeError('Graph contains components without root nodes.')

        if callback(stack[stack_top], np.nan, False):
            # _log.debug(debug_message.format(timer.total()))
            return

        visited[stack[stack_top]] = True

        # Perform depth-first traversal of this connected component
        while stack_top >= 0:
            # Pop a previously visited node from the stack
            node_pre = stack[stack_top]
            stack_top -= 1

            # Visit all adjacent nodes
            idx_src = e_src == node_pre
            idx_dst = e_dst == node_pre
            adjacent_nodes = np.concatenate((e_dst[idx_src], e_src[idx_dst]))

            for node in adjacent_nodes:
                if callback(node, node_pre, visited[node]):
                    # _log.debug(debug_message.format(timer.total()))
                    return

            # Push previously unvisited adjacent nodes to the top of the stack
            adjacent_nodes = adjacent_nodes[~visited[adjacent_nodes]]
            new_top = stack_top + len(adjacent_nodes)
            stack[(stack_top + 1):(new_top + 1)] = adjacent_nodes[::-1]
            stack_top = new_top
            visited[adjacent_nodes] = True

            # Remove traversed edges
            idx = idx_src | idx_dst
            e_src[idx] = -1
            e_dst[idx] = -1

    # _log.debug(debug_message.format(timer.total()))


def is_acyclic_graph(nodes, edges):
    """Return ``True`` if the graph is acyclic and ``False`` otherwise."""

    # Flag for acyclicity: For the closure to work, we have to avoid a
    # direct assignment in the callback as this causes Python to create a
    # local variable. Due to this, the flag is stored in a list.
    acyclic = [True]

    def detect_cycles(node, node_pre, cycle):  # pylint: disable=unused-argument
        """Callback for the graph traversal to detect cycles."""
        if cycle:
            acyclic[0] = False
            return True  # Abort traversal
        return False

    traverse_graph(nodes, edges, detect_cycles)
    return acyclic[0]


def get_graph_components(nodes, edges, roots=None):
    """
    Return a list of connected components.

    The returned list contains an element for every connected component, where
    the element constitutes an array with all nodes of that component. The
    first element of the array is the root node of the respective component.
    """
    components = []

    def identify_components(node, node_pre, cycle):
        """Callback for the graph traversal to identify components."""
        if np.isnan(node_pre):
            components.append([node])
        elif not cycle:
            components[-1].append(node)

    traverse_graph(nodes, edges, identify_components, roots)
    return [np.array(component) for component in components]


def get_minimum_spanning_tree(edges, weights):
    """Return the minimum spanning tree in terms of an edges tuple."""
    (e_src, e_dst) = edges

    if not e_src.size:
        return np.ndarray(0, dtype=hynet_int_), np.ndarray(0, dtype=hynet_int_)

    if min([min(e_src), min(e_dst)]) < 0:
        raise RuntimeError('Negative node numbers are not supported.')

    N = max([max(e_src), max(e_dst)]) + 1
    A = create_sparse_matrix(e_src, e_dst, weights, N, N, dtype=hynet_float_)
    A += A.transpose()
    return scipy.sparse.csgraph.minimum_spanning_tree(A).nonzero()



"""
Visualization of network graphs.

*hynet* uses NetworkX to export graph data. NetworkX can export to various
popular formats, including the GraphViz DOT format and JSON.
"""

import copy
import json
import logging
from math import isnan, isinf

import networkx as nx
from networkx.readwrite import json_graph

from hynet.scenario.representation import Scenario
from hynet.model.steady_state import SystemModel
from hynet.opf.result import OPFResult

_log = logging.getLogger(__name__)


def create_networkx_graph(data):
    """
    Return a NetworkX graph object for the provided network graph.

    Buses and injectors are inserted as nodes. Branches and converters are
    inserted as edges between the buses, while injectors are connected via
    additional edges to their respective terminal bus. The scenario data and,
    if provided, the OPF result data of the individual entities is added as
    attributes to the respective graph elements. The returned NetworkX graph
    can be exported into various formats, please refer to the documentation of
    NetworkX.

    Parameters
    ----------
    data : Scenario or SystemModel or OPFResult
        ``Scenario`` object, ``SystemModel`` object, or ``OPFResult`` that
        contains the network graph information.

    Returns
    -------
    graph : nx.Graph
        NetworkX graph representation of the grid's network graph.

    See Also
    --------
    export_networkx_graph_to_json
    """
    graph = nx.Graph()
    if isinstance(data, Scenario):
        scenario = data
    elif isinstance(data, (SystemModel, OPFResult)):
        scenario = data.scenario
    else:
        raise ValueError(("The argument 'data' must be a Scenario object, "
                          "a SystemModel object, or an OPF result."))

    bus_node = 'Bus {0}'
    injector_node = 'Injector {0}'

    buses = _restore_index(scenario.bus)
    branches = _restore_index(scenario.branch)
    converters = _restore_index(scenario.converter)
    injectors = _restore_index(scenario.injector)

    if isinstance(data, OPFResult):
        buses = buses.join(data.bus, rsuffix='_opf')
        branches = branches.join(data.branch, rsuffix='_opf')
        converters = converters.join(data.converter, rsuffix='_opf')
        injectors = injectors.join(data.injector, rsuffix='_opf')

    buses = buses.to_dict(orient='records')
    branches = branches.to_dict(orient='records')
    converters = converters.to_dict(orient='records')
    injectors = injectors.to_dict(orient='records')

    for bus in buses:
        bus['graph_type'] = 'bus'
        graph.add_node(bus_node.format(bus['id']), attr_dict=bus)

    for branch in branches:
        branch['graph_type'] = 'branch'
        graph.add_edge(bus_node.format(branch['src']),
                       bus_node.format(branch['dst']),
                       attr_dict=branch)

    for converter in converters:
        converter['graph_type'] = 'converter'
        graph.add_edge(bus_node.format(converter['src']),
                       bus_node.format(converter['dst']),
                       attr_dict=converter)

    for injector in injectors:
        injector['graph_type'] = 'injector'
        graph.add_node(injector_node.format(injector['id']),
                       attr_dict=injector)
        graph.add_edge(injector_node.format(injector['id']),
                       bus_node.format(injector['bus']),
                       attr_dict={'graph_type': 'terminal',
                                  'injector': injector['id'],
                                  'bus': injector['bus']})

    return graph


def export_networkx_graph_to_json(graph, output_file):
    """
    Exports a NetworkX graph object to a JSON file.

    The specified network graph is exported to the JSON format, which is
    supported by D3 [1]_.

    **Remark:** As a simple tool to visualize the graph, this subpackage
    includes the web page ``show_graph.html`` in the subdirectory
    ``rendering``. To use it, store the graph object to the file
    ``network_graph.json`` and open the aforementioned HTML file.

    Parameters
    ----------
    graph: nx.Graph
        The graph that shall be exported to the JSON format.
    output_file: str
        The file name to which the JSON data shall be exported.

    See Also
    --------
    create_networkx_graph

    References
    ----------
    .. [1] https://github.com/d3/d3
    """
    json_output = json_graph.node_link_data(graph)
    json_output = _clear_invalid_json_values(json_output)

    with open(output_file, 'w') as file:
        json.dump(json_output, file,
                  default=lambda x: x.__str__(),
                  allow_nan=False)


def _clear_invalid_json_values(dictionary):
    """
    Convert all ``nan`` and ``inf`` values in the dictionary to ``None``.

    This function ensures that there are no not-a-number and infinity values
    in the input dictionary, recursively. This avoids that some parsers throw
    errors when attempting to import the data.

    Parameters
    ----------
    dictionary: dict
        A dictionary, potentially containing invalid JSON values.

    Returns
    -------
    dictionary: dict
        A dictionary that has all invalid JSON values replaced by ``None``.
    """
    for k, v in dictionary.items():
        if isinstance(v, dict):
            _clear_invalid_json_values(v)
        if isinstance(v, list):
            for x in v:
                _clear_invalid_json_values(x)
        elif isinstance(v, float):
            if isnan(v) or isinf(v):
                dictionary[k] = None
    return dictionary


def _restore_index(dataframe):
    """
    Returns a copy of the data frame with its index in the column ``id``.

    Parameters
    ----------
    dataframe: pd.DataFrame
        Input data frame.

    Returns
    -------
    dataframe_copy: pd.DataFrame
        Deep copy of the data frame with an additional column ``id`` that
        contains the data frame's index.
    """
    dataframe_copy = copy.deepcopy(dataframe)
    dataframe_copy['id'] = dataframe_copy.index
    return dataframe_copy



"""
Network reduction to a "copper plate".
"""

import logging
import copy

import numpy as np

from hynet.types_ import BusType
from hynet.utilities.base import Timer
from hynet.scenario.representation import Scenario
from hynet.scenario.capability import CapRegion

_log = logging.getLogger(__name__)


def reduce_to_copper_plate(scenario):
    """
    Return a deep copy of the scenario with the grid reduced to a copper plate.

    In the "copper plate" reduction, every connected grid is reduced to a
    single bus, i.e., the impact of the power grid is neglected and the network
    laws are reduced to simply balance the total injection and total load.
    """
    timer = Timer()
    cpr = Scenario()
    cpr.id = scenario.id
    cpr.name = scenario.name
    cpr.time = scenario.time
    cpr.database_uri = scenario.database_uri
    cpr.grid_name = "COPPER PLATE REDUCTION"
    cpr.base_mva = scenario.base_mva
    cpr.bus = copy.deepcopy(scenario.bus)
    cpr.injector = copy.deepcopy(scenario.injector)

    for island in scenario.get_islands():
        # Define the supplementary bus to which this island is reduced
        ref_bus = island[0]

        # Accumulate all active load at the supplementary bus
        cpr.bus.at[ref_bus, 'load'] = cpr.bus.loc[island, 'load'].real.sum()

        # Eliminate all other buses of the island and init. the supp. bus
        cpr.bus.drop(island[1:], axis='index', inplace=True)
        cpr.bus.at[ref_bus, 'type'] = BusType.DC
        cpr.bus.at[ref_bus, 'ref'] = True
        cpr.bus.at[ref_bus, 'base_kv'] = np.nan
        cpr.bus.at[ref_bus, 'y_tld'] = 0
        cpr.bus.at[ref_bus, 'v_min'] = 1.0
        cpr.bus.at[ref_bus, 'v_max'] = 1.0
        cpr.bus.at[ref_bus, 'annotation'] = ''

        # Connect all injectors of this island to the supplementary bus,
        # eliminate reactive power capability, and deep copy if necessary
        index = cpr.injector.index[cpr.injector['bus'].isin(island)]
        cpr.injector.loc[index, 'bus'] = ref_bus
        for id_ in index:
            cap = cpr.injector.at[id_, 'cap']
            cpr.injector.at[id_, 'cap'] = CapRegion([cap.p_min, cap.p_max])
            cpr.injector.at[id_, 'cost_p'] = \
                copy.deepcopy(cpr.injector.at[id_, 'cost_p'])
            cpr.injector.at[id_, 'cost_q'] = None

    _log.debug("Copper plate reduction ({:.3f} sec.)".format(timer.total()))
    return cpr



#pylint: disable=missing-docstring,redefined-outer-name,invalid-name,broad-except
"""
*hynet* command line tool.
"""

import argparse

import hynet as ht


def command_test(args):  # pylint: disable=unused-argument
    if ht.test_installation():
        print("\nAll installed solvers work properly.")
    else:
        print("\nTest failed. Please check the installation.")
        raise SystemExit(1)


def command_import(args):
    try:
        output_file = \
            ht.import_matpower_test_case(input_file=args.file_name,
                                         output_file=args.output,
                                         grid_name=args.grid_name,
                                         description=args.description,
                                         num_sample_points=args.num_sample_points,
                                         res_detection=args.res_detection)
    except Exception as exception:
        print("ERROR: " + str(exception))
        print("Failed to import '{:s}'.".format(args.file_name))
        raise SystemExit(1)
    else:
        print("Successfully imported to '{:s}'.".format(output_file))


def command_client(args):
    if args.num_workers is not None and args.num_workers > 1:
        ht.config.GENERAL['parallelize'] = False
    try:
        ht.start_optimization_client(server_ip=args.server_ip,
                                     port=args.port,
                                     authkey=args.authkey,
                                     num_workers=args.num_workers)
    except KeyboardInterrupt:
        pass  # Terminated by the user
    except Exception as exception:
        print("ERROR: " + str(exception))
        raise SystemExit(1)


def command_opf(args):
    if args.type:
        args.type = ht.SolverType(args.type)
    try:
        print(ht.calc_opf(ht.connect(args.database),
                          scenario_id=args.scenario,
                          solver_type=args.type))
    except Exception as exception:
        print("ERROR: " + str(exception))
        raise SystemExit(1)


parser = argparse.ArgumentParser(
    description="An optimal power flow framework for hybrid AC/DC power systems.",
    prog='python -m hynet')
parser.add_argument(
    '-v', '--version',
    action='store_true',
    help="show the program version and exit")
subparsers = parser.add_subparsers()

parser_test = subparsers.add_parser(
    'test',
    help='test the installation')
parser_test.set_defaults(execute_command=command_test)

parser_imp = subparsers.add_parser(
    'import',
    help="import data into the hynet grid database format")
parser_imp.add_argument(
    'file_name',
    type=str,
    help="MATPOWER test case as MATLAB MAT-file with 'mpc'")
parser_imp.add_argument(
    '-o', '--output',
    type=str,
    help="output file name")
parser_imp.add_argument(
    '-g', '--grid_name',
    type=str,
    default=None,
    help="name of the grid")
parser_imp.add_argument(
    '-d', '--description',
    type=str,
    default='',
    help="description and copyright/licensing information")
parser_imp.add_argument(
    '-n', '--num_sample_points',
    type=int,
    default=10,
    help="sample points for polynomial to PWL conversion")
parser_imp.add_argument(
    '-r', '--res_detection',
    action='store_true',
    help="enable RES injector type detection")
parser_imp.set_defaults(execute_command=command_import)

parser_client = subparsers.add_parser(
    'client',
    help="start a hynet optimization client on this machine")
parser_client.add_argument(
    'server_ip',
    type=str,
    help="IP address of the hynet optimization server")
parser_client.add_argument(
    '-p', '--port',
    type=int,
    default=None,
    help="TCP port number of the hynet optimization server")
parser_client.add_argument(
    '-a', '--authkey',
    type=str,
    default=None,
    help="authentication key for the hynet optimization server")
parser_client.add_argument(
    '-n', '--num_workers',
    type=int,
    default=None,
    help="number of worker processes that should run in parallel")
parser_client.set_defaults(execute_command=command_client)

parser_opf = subparsers.add_parser(
    'opf',
    help="calculate an optimal power flow")
parser_opf.add_argument(
    'database',
    type=str,
    help="file name of the grid database")
parser_opf.add_argument(
    '-s', '--scenario',
    metavar='ID',
    type=int,
    default=0,
    help="specify the scenario ID")
parser_opf.add_argument(
    '-t', '--type',
    type=str.upper,
    choices=['QCQP', 'SDR', 'SOCR'],
    help="specify the solver type")
parser_opf.set_defaults(execute_command=command_opf)

args = parser.parse_args()

if args.version:
    print("hynet v" + ht.__version__)
elif hasattr(args, 'execute_command'):
    try:
        args.execute_command(args)
    except Exception as exception:
        print(exception)
else:
    parser.print_usage()



"""
CVXPY-based solver for the *hynet*-specific QCQP problem (SD and SOC relaxation).

**Caution:** In our experiments, we observed bad performance and, especially
in case of the SDR, convergence issues. Furthermore, we encountered issues
with CVXPY on Windows machines. Due to this, this solver interface is not
supported officially.
"""

import logging
import abc

import cvxpy as cvx
import numpy as np

from hynet.types_ import hynet_float_, SolverType, SolverStatus
from hynet.utilities.base import Timer
from hynet.qcqp.solver import SolverInterface
from hynet.qcqp.problem import QCQPPoint
from hynet.qcqp.result import QCQPResult

_log = logging.getLogger(__name__)


def parse_status(status):
    """Parse the CVXPY status into a ``SolverStatus``."""
    if status.upper() == 'OPTIMAL':
        return SolverStatus.SOLVED
    if status.upper() in ['OPTIMAL_INACCURATE',
                          'UNBOUNDED_INACCURATE',
                          'INFEASIBLE_INACCURATE']:
        return SolverStatus.INACCURATE
    if status.upper() == 'INFEASIBLE':
        return SolverStatus.INFEASIBLE
    if status.upper() == 'UNBOUNDED':
        return SolverStatus.UNBOUNDED
    return SolverStatus.FAILED


class SolverBase(SolverInterface):
    """Base class for CVXPY-based solvers for a relaxation of the QCQP."""

    def __init__(self, solver_name='CVXOPT', **kwds):
        super().__init__(**kwds)
        self.solver_name = solver_name

        if self.solver_name.upper() == 'CVXOPT':
            if not 'abstol' in self.param:
                self.param['abstol'] = 1e-6
            if not 'reltol' in self.param:
                self.param['reltol'] = 1e-5
            if not 'feastol' in self.param:
                self.param['feastol'] = 1e-6
            if not 'max_iters' in self.param:
                self.param['max_iters'] = 500
            if not 'kktsolver' in self.param:
                # With the default ('chol', i.e., a Cholesky factorization for
                # preprocessing), CVXPY frequently fails with an exception,
                # where the log states "Terminated (singular KKT matrix)."
                self.param['kktsolver'] = 'robust'

    @staticmethod
    @abc.abstractmethod
    def _add_voltage_matrix_constraints(constraint_list, qcqp, x):
        """Add the SOC or PSD constraint(s) on the ``v_tld``-part of ``x``."""

    def solve(self, qcqp):
        """
        Solve the given QCQP via a relaxation using CVXPY.

        The relaxation-specific constraints on the bus voltage matrix are
        added via the (abstract) method ``_add_voltage_matrix_constraints``,
        which must be implemented in a derived class.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the relaxed problem associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                ...and SOC or PSD constraint(s) on v_tld-part of x
        #
        # For the following code, I used the CVXPY documentation provided at
        # http://www.cvxpy.org/tutorial/index.html

        timer = Timer()
        # ------------------ generate data --------------------------
        (c, A, b, B, d, x_lb, x_ub) = qcqp.get_vectorization()

        # -------------------- size of the data ---------------------
        dim_x = c.shape[0]

        # -------------- create optimization variable ---------------
        x = cvx.Variable(dim_x)

        # ---- convert SciPy sparse matrices to CVXPY constants -----
        # (See the caveat under "Can I use SciPy sparse matrices with CVXPY?"
        #  in the CVXPY FAQ at https://www.cvxpy.org/faq/index.html)
        c = cvx.Constant(c)
        A = cvx.Constant(A)
        B = cvx.Constant(B)

        # ------------------- add the constraints -------------------
        constraint_list = [A * x == b,
                           B * x <= d]

        # Add the specified lower and upper bounds
        idx_lb = ~np.isnan(x_lb)
        constraint_list.append(x[idx_lb] >= x_lb[idx_lb])

        idx_ub = ~np.isnan(x_ub)
        constraint_list.append(x[idx_ub] <= x_ub[idx_ub])

        # Add relaxation-specific constraints on the voltage matrix
        # (This method is overridden appropriately in derived classes)
        self._add_voltage_matrix_constraints(constraint_list, qcqp, x)

        # -------------------- set the objective --------------------
        obj = cvx.Minimize(c.T*x)

        # -------------------- create a problem ---------------------
        prob = cvx.Problem(obj, constraint_list)

        _log.debug(self.type.name + " ~ Modeling ({:.3f} sec.)"
                   .format(timer.interval()))

        # ----------- optimization and solution recovery ------------
        prob.solve(solver=self.solver_name.upper(),
                   verbose=self.verbose,
                   **self.param)
        solver_time = timer.interval()
        _log.debug(self.type.name + " ~ Solver ({:.3f} sec.)"
                   .format(solver_time))

        status = parse_status(prob.status)

        if status in [SolverStatus.SOLVED, SolverStatus.INACCURATE]:
            x = np.asarray(x.value).reshape(dim_x)

            (V, f, s, z) = qcqp.split_vectorization_optimizer(x)
            v = self.rank1approx(V, qcqp.edges, qcqp.roots)
            optimizer = QCQPPoint(v, f, s, z)

            dv_eq = np.asarray(constraint_list[0].dual_value).reshape(-1)
            dv_ineq = np.asarray(constraint_list[1].dual_value).reshape(-1)

            dv_lb = np.nan * np.ones(dim_x, dtype=hynet_float_)
            dv_lb[idx_lb] = constraint_list[2].dual_value
            dv_lb = qcqp.split_vectorization_bound_dual(dv_lb)

            dv_ub = np.nan * np.ones(dim_x, dtype=hynet_float_)
            dv_ub[idx_ub] = constraint_list[3].dual_value
            dv_ub = qcqp.split_vectorization_bound_dual(dv_ub)

            optimal_value = prob.value

            result = QCQPResult(qcqp, self, status, solver_time,
                                optimizer=optimizer,
                                V=V,
                                optimal_value=optimal_value,
                                dv_lb=dv_lb,
                                dv_ub=dv_ub,
                                dv_eq=dv_eq,
                                dv_ineq=dv_ineq)
        else:
            result = QCQPResult(qcqp, self, status, solver_time)

        _log.debug(self.type.name + " ~ Result creation ({:.3f} sec.)"
                   .format(timer.interval()))
        _log.debug(self.type.name + " ~ Total time for solver ({:.3f} sec.)"
                   .format(timer.total()))
        return result


class SOCRSolver(SolverBase):
    """
    CVXPY-based solver for a second-order cone relaxation of the QCQP.

    For information on solver-specific parameters, see e.g. [1]_. To retrieve a
    list of available solvers, use

    >>> import cvxpy
    >>> cvxpy.installed_solvers()

    By default, CVXOPT is selected. For appropriate performance in OPF
    calculations, its default parameters are adapted. Access the ``param``
    attribute of an object for their inspection.

    **Caution:** We observed that ECOS sometimes yields inaccurate dual variables.

    See Also
    --------
    hynet.qcqp.solver.SolverInterface : Abstract base class for QCQP solvers.

    References
    ----------
    .. [1] http://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options
    """

    @property
    def name(self):
        return "CVXPY + " + self.solver_name.upper()

    @property
    def type(self):
        return SolverType.SOCR

    def solve(self, qcqp):  # pylint: disable=useless-super-delegation
        """
        Solve the given QCQP via its second-order cone relaxation using CVXPY.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the SOCR associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                   ...and SOC constraints on v_tld-part of x
        return super().solve(qcqp)

    @staticmethod
    def _add_voltage_matrix_constraints(constraint_list, qcqp, x):
        """Add the SOC constraints on the ``v_tld``-part of ``x``."""
        # Second-order cone constraints on V (cf. QCQP.get_vectorization)
        #
        # PSD constraint on 2x2 principal submatrix =>
        #  (a) diagonal is nonnegative -> ensured by LB on voltage mag.
        #  (b) determinant is nonnegative -> rotated SOC:
        #
        #               V_ij * conj(V_ij) <= V_ii * V_jj
        #                             <=>
        #          real(V_ij)^2 + imag(V_ij)^2 <= V_ii * V_jj
        #
        #      where i is the source bus and j is the destination bus
        #      of the considered edge. The latter is implemented below.
        #
        # The rotated SOC specification is not supported by CVXPY. Due to
        # this, we reformulate it as a standard SOC constraint:
        #
        #               V_ij * conj(V_ij) <= V_ii * V_jj
        #
        #                             <=>
        #
        #           || 2 * Re(V_ij) ||^2
        #           || 2 * Im(V_ij) ||   <= (V_ii + V_jj)^2
        #           || V_ii - V_jj  ||_2
        #
        #                             <=> (with (a) satisfied)
        #
        #               [ 2 * Re(V_ij) ]
        #               [ 2 * Im(V_ij) ] <= V_ii + V_jj
        #               [ V_ii - V_jj  ]
        dim_v = qcqp.dim_v
        K = qcqp.num_edges
        (e_src, e_dst) = qcqp.edges

        for k in range(K):
            constraint_list.append(
                cvx.norm(cvx.vstack([2*x[dim_v + k],
                                     2*x[dim_v + k + K],
                                     x[e_src[k]] - x[e_dst[k]]
                                     ]), 2) <= x[e_src[k]] + x[e_dst[k]])


class SDRSolver(SolverBase):
    """
    CVXPY-based solver for a semidefinite relaxation of the QCQP.

    **Caution:** As we experienced convergence and accuracy issues with the
    default solver CVXOPT, we do not include this solver interface officially.

    For information on solver-specific parameters, see e.g. [1]_. To retrieve a
    list of available solvers, use

    >>> import cvxpy
    >>> cvxpy.installed_solvers()

    By default, CVXOPT is selected. For appropriate performance in OPF
    calculations, its default parameters are adapted. Access the ``param``
    attribute of an object for their inspection.

    See Also
    --------
    hynet.qcqp.solver.SolverInterface : Abstract base class for QCQP solvers.

    References
    ----------
    .. [1] http://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options
    """

    @property
    def name(self):
        return "CVXPY + " + self.solver_name.upper()

    @property
    def type(self):
        return SolverType.SDR

    def solve(self, qcqp):  # pylint: disable=useless-super-delegation
        """
        Solve the given QCQP via its semidefinite relaxation using CVXPY.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the SDR associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                   ...and a PSD constraint on the v_tld-part of x
        return super().solve(qcqp)

    @staticmethod
    def _add_voltage_matrix_constraints(constraint_list, qcqp, x):
        """Add the PSD constraint on the ``v_tld``-part of ``x``."""
        # PSD constraint on V
        #
        # Consider the splitting of the Hermitian matrix V into its
        # real and imaginary part, i.e., V = V_r + j*V_i. Then,
        #
        #       V is PSD    <=>    V_cr = [ V_r  -V_i ] is PSD
        #                                 [ V_i   V_r ]
        #
        # This code implements the real-valued equivalent of V >= 0,
        # i.e., V_cr >= 0. To this end, a semidefinite matrix variable
        # V_cr is added and all relevant elements are equated to v_tld.
        dim_v = qcqp.dim_v
        K = qcqp.num_edges
        (e_src, e_dst) = qcqp.edges

        if np.any(e_src >= e_dst):
            raise RuntimeError("For simplicity, this solver assumes "
                               "that the source node index is smaller "
                               "than the destination node index in the "
                               "sparsity specification. At the time of "
                               "implementation, this was the default.")

        # Note that previous to CVXPY v1.0, the syntax to define a symmetric
        # and positive semidefinite matrix was
        #   V_tld = cvx.Semidef(2 * dim_v)
        # Starting with CVXPY v1.0, it is defined as:
        V_tld = cvx.Variable((2 * dim_v, 2 * dim_v), PSD=True)

        for k in range(K):     # Equating of off-diagonal elements
            i = e_src[k]
            j = e_dst[k]
            constraint_list.append(V_tld[i, j] == x[dim_v + k])
            constraint_list.append(V_tld[i + dim_v, j + dim_v] == x[dim_v + k])
            constraint_list.append(V_tld[i + dim_v, j] == x[dim_v + k + K])
            constraint_list.append(V_tld[j + dim_v, i] == -x[dim_v + k + K])

        for i in range(dim_v):  # Equating of diagonal elements
            constraint_list.append(V_tld[i, i] == x[i])
            constraint_list.append(V_tld[i + dim_v, i + dim_v] == x[i])
            constraint_list.append(V_tld[i + dim_v, i] == 0)



"""
    This module contains various classes to provide interchangable data types
    between the view classes
"""

from collections import namedtuple
from enum import Enum


class Point(namedtuple('Point', ['p', 'q'])):
    """Representation of a point in the P/Q-plane"""

    def __str__(self):
        return f'(p:{self.p}, q:{self.q})'


class LinearFunction:
    """Helper class to simplify the handling of the PWLs"""

    def __init__(self, halfspace, limits):
        self.slope = halfspace.slope
        self.offset = halfspace.offset
        self.t = (limits.q * halfspace.offset - self.slope * limits.p)
    # pylint: disable=invalid-name
    def f(self, p):
        """
            This function computes the PWL function for the input value p.

        Parameters
        ----------
        p: float
            The active power value for this linear function

        Returns
        -------
            The result of the linear curve equation for the input value:
                $\text{self.slope}*p + \text{self.t}$
        """
        return self.slope * p + self.t

    def f_1(self, q):
        """
            This function computes the inverse PWL function for the input
            value q.

            Parameters
            ----------
            q: float
                The reactive power value for this linear function.

            Returns
            -------
            The result of the linear curve equation for the input value:
                        $ (q - \text{self.t}) / \text{self.slope}$
        """

        return (q - self.t) / self.slope

    def intersect_with(self, other):
        """
            This function computes the intersection of this LinearFunction with
            the LinearFunction `other`.

            Parameters
            ----------
            other: LinearFunction
                The other LinearFunction we want to compute the intersection with.

            Returns
            -------
                Either None or a Point that represents the intersection of the two
                instances of LinearFunction
        """
        denominator = other.slope - self.slope
        if denominator == 0:
            return None
        p = (self.t - other.t) / denominator
        q = other.slope * p + other.t
        return Point(p, q)

    def __repr__(self):
        return f"<f(p)={self.slope}p + {self.t}, f(q) = " \
               f"(q - {self.t}) / {self.slope}>"


class AutoValue(Enum):
    """
        This class provides a base class for enum that have a predefined,
        readable string representation.
    """
    def __repr__(self):
        return '<%s.%s>' % (self.__class__.__name__, self.name)


class Axis(AutoValue):
    """
        This enum represents the two axes.
    """
    ACTIVE_POWER = 1
    REACTIVE_POWER = 2


class Bound(AutoValue):
    """
        This enum represents the two bounds of the reactive and active limits
        of a capability region.
    """
    MIN = 1
    MAX = 2


class Orientation(AutoValue):
    """
        This enum represents the four possibilities for halfspaces. These
        should only appear in pairs of two: {RIGHT, LEFT} x {TOP, BOTTOM}.
    """
    RIGHT = 1
    LEFT = 2
    TOP = 3
    BOTTOM = 4



"""
Representation of (injector) cost functions in *hynet*.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d

from hynet.types_ import hynet_float_


class PWLFunction:
    """
    Representation of a piecewise linear function ``f: R -> R``.

    Parameters
    ----------
    samples : (numpy.ndarray[.hynet_float_], numpy.ndarray[.hynet_float_]), optional
        Tuple ``(x, y)`` of ``x``- and ``y``-coordinates of sample points of
        the piecewise linear function, i.e., ``(x[0], y[0]), ..., (x[N], y[N])``.
    """

    def __init__(self, samples=None):
        """Create the PWL function with sample points 'samples'"""
        self._x = np.array([0, 1], dtype=hynet_float_)
        self._y = np.array([0, 0], dtype=hynet_float_)
        self.samples = ((0, 1), (0, 0)) if samples is None else samples

    @property
    def samples(self):
        """
        Return the tuple ``(x, y)`` of ``x``- and ``y``-coordinate arrays.
        """
        return self._x, self._y

    @samples.setter
    def samples(self, value):
        """
        Set the tuple ``(x, y)`` of ``x``- and ``y``-coordinate arrays.
        """
        if len(value) != 2:
            raise ValueError("Expecting a tuple (x, y) of coordinates.")
        x = np.array(value[0], dtype=hynet_float_)
        y = np.array(value[1], dtype=hynet_float_)
        if len(x) < 2 or len(x) != len(y) or x.ndim != 1 or y.ndim != 1:
            raise ValueError("Expecting a tuple (x, y) of coordinates "
                             "comprising at least two points.")
        if any(np.diff(x) <= 0):
            raise ValueError("The x-coordinate of samples must be "
                             "strictly increasing.")
        (self._x, self._y) = (x, y)

    def __repr__(self):
        if len(self._x) == 2:
            return str(np.asscalar(np.diff(self._y) / np.diff(self._x)))
        return '-'.join(['({0:g},{1:g})'.format(x, y)
                         for x, y in zip(self._x, self._y)])

    def __eq__(self, other):
        """Return True if the cost functions feature the same sample points."""
        if other is None:
            return False
        return (np.array_equal(self._x, other.samples[0]) and
                np.array_equal(self._y, other.samples[1]))

    def evaluate(self, x):
        """
        Evaluate the function at ``x``, i.e., return ``y = f(x)``.
        """
        result = interp1d(self._x, self._y,
                          kind='linear', fill_value='extrapolate')(x)
        if np.isscalar(x):
            return np.asscalar(result)
        return result

    def show(self):
        """
        Show the piecewise linear function.
        """
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self._x, self._y,
                color='xkcd:sea blue', linestyle='-', linewidth=1)
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        fig.tight_layout()
        fig.show()
        return fig

    def scale(self, scaling):
        """
        Scale the function by ``scaling``, i.e., ``f(x) -> scaling * f(x)``.
        """
        if scaling < 0:
            raise ValueError("Expecting a nonnegative scaling factor.")
        self._y *= scaling
        return self

    def is_convex(self):
        """
        Return ``True`` if the function is convex.
        """
        if len(self._x) == 2:
            # Affine function
            return True
        # Convex if curvature is nonnegative
        return all(np.diff(np.diff(self._y) / np.diff(self._x)) >= 0)

    def get_epigraph_polyhedron(self):
        """
        Return ``(A, b)`` such that ``f(x) = min{z in R: z*1 >= A*x + b}``.

        Note that ``f`` must be **convex**.
        """
        A = np.diff(self._y) / np.diff(self._x)        # Slope
        b = self._y[1:] - np.multiply(A, self._x[1:])  # Offset
        return A, b



#pylint: disable=wrong-import-position,invalid-name
"""
*hynet:* An optimal power flow framework for hybrid AC/DC power systems.

For more information, please refer to ``README.md``, which is provided
alongside *hynet*, as well as the docstrings of the individual classes and
functions. The variables below are set up during package initialization.

Parameters
----------
AVAILABLE_SOLVERS : list
    List of classes for all solvers available on the current system.
__version__ : str
    *hynet* version.
"""

__version__ = '1.0.3'  # Update the version in setup.py and conf.py as well

import sys as _sys

if _sys.version_info < (3, 5):
    raise ImportError('hynet requires Python 3.5 or higher.')


from hynet.types_ import (hynet_id_,
                          hynet_int_,
                          hynet_float_,
                          hynet_complex_,
                          hynet_eps,
                          BusType,
                          BranchType,
                          InjectorType,
                          SolverType,
                          SolverStatus)

from hynet.data.connection import connect
from hynet.data.interface import (get_scenario_info,
                                  load_scenario,
                                  save_scenario,
                                  initialize_database,
                                  remove_scenarios,
                                  copy_scenarios)
from hynet.data.import_ import import_matpower_test_case

from hynet.scenario.capability import HalfSpace, CapRegion, ConverterCapRegion
from hynet.scenario.cost import PWLFunction
from hynet.scenario.representation import Scenario

from hynet.model.steady_state import SystemModel

from hynet.opf.calc import (calc_opf,
                            select_solver,
                            get_relaxation_initial_point,
                            get_copper_plate_initial_point)

from hynet.solver import AVAILABLE_SOLVERS

from hynet.test.installation import test_installation

from hynet.visual.opf import (show_lmp_profile,
                              show_voltage_profile,
                              show_branch_flow_profile,
                              show_ampacity_dual_profile,
                              show_converter_flow_profile,
                              show_dispatch_profile)

from hynet.utilities.base import Timer

from hynet.distributed.server import start_optimization_server, OptimizationJob
from hynet.distributed.client import start_optimization_client



"""
Rank-1 approximation of a partial Hermitian matrix.
"""

import logging

import numpy as np
import matplotlib.pyplot as plt

from hynet.types_ import hynet_float_, hynet_complex_
from hynet.utilities.base import (create_sparse_matrix,
                                  create_sparse_diag_matrix,
                                  Timer)
from hynet.utilities.graph import traverse_graph, get_graph_components

_log = logging.getLogger(__name__)


def rank1approx_via_traversal(V, edges, roots):
    """
    Return a rank-1 approximation ``vv^H`` of ``V`` based on a graph traversal.

    Assuming that ``V = vv^H`` (i.e., ``rank(V) = 1``), the off-diagonal
    element in row ``i`` and column ``j`` is ``V_ij = v_i*conj(v_j)``. Thus,
    ``|v_j| = sqrt(V_jj)`` and ``arg(v_j) = arg(v_i) - arg(V_ij)``. This is
    utilized to recover ``v`` by setting its element's magnitude to the square
    root of the diagonal elements of ``V`` and reconstructing the angle of its
    elements by accumulating the angle differences from the respective root
    node of in the graph associated with ``V``. The angle at the root node(s)
    is set to zero.
    """
    V = V.tocsr()  # Conversion to CSR for indexing
    dim_v = V.shape[0]
    v = np.zeros(dim_v, dtype=hynet_complex_)
    np.sqrt(V.diagonal().real, out=v)

    def update_angle(node, node_pre, cycle):
        """Callback for the graph traversal to update the angles."""
        if cycle or np.isnan(node_pre):
            return
        angle = np.angle(v[node_pre]) - np.angle(V[node_pre, node])
        v[node] *= np.exp(1j * angle)

    traverse_graph(np.arange(dim_v), edges, update_angle, roots)
    return v


def calc_rank1approx_rel_err(V, v, edges):
    """Calculate the relative reconstruction error for all edges."""
    (e_src, e_dst) = edges
    if not e_src.size:
        return np.ndarray(0, dtype=hynet_float_)
    vv_ = np.multiply(v[e_src], v[e_dst].conj())
    V_ = np.asarray(V.tocsr()[e_src, e_dst])[0]
    return np.true_divide(np.abs(vv_ - V_), np.abs(vv_))


def calc_rank1approx_mse(V, v, edges):
    """
    Calculate the mean squared error for the rank-1 approximation.

    Returns the mean of the squared error of the elements on the diagonal as
    well as those on sparsity pattern defined by ``edges``.
    """
    # ATTENTION: Use caution when modifying this function as the least
    # squares rank-1 approximation depends on this particular MSE calculation.
    (e_src, e_dst) = edges
    vv_ = np.concatenate((np.square(np.abs(v)),
                          np.multiply(v[e_src], v[e_dst].conj()),
                          np.multiply(v[e_dst], v[e_src].conj())))
    # The case distinction is necessary due to inconsistencies of SciPy...
    V = V.tocsr()  # Conversion to CSR for indexing
    V_ = V.diagonal()
    if e_src.size:
        V_ = np.concatenate((V_,
                             np.asarray(V[e_src, e_dst])[0],
                             np.asarray(V[e_dst, e_src])[0]))
    return np.square(np.abs(vv_ - V_)).mean()


def rank1approx_via_least_squares(V, edges, roots, grad_thres=1e-7,
                                  mse_rel_thres=0.002, max_iter=300,
                                  show_convergence_plot=False):
    """
    Return a rank-1 approximation ``vv^H`` of ``V`` by minimizing the squared error.

    The rank-1 approximation may be put as the following optimization problem::

        minimize  ||P(vv^H - V)||_F^2
        v in C^N

    Therein, ``P(X)`` is the projection of the matrix ``X`` onto the sparsity
    pattern defined by ``edges``, i.e., all diagonal elements and the
    off-diagonal elements ``(edges[k][0], edges[k][1])`` and
    ``(edges[k][1], edges[k][0])`` for ``k = 0,...,K-1``, where ``K`` is the
    number of edges. ``||.||_F`` denotes the Frobenius norm. For ``V >= 0``
    (psd) and ``V != 0``, this problem is nonconvex, i.e., the objective is not
    convex over the entire ``C^N``. (If ``rank(V) = 1`` and ``V = xx^H``, then
    the objective is locally convex at ``x``.)

    This function finds a (local) optimizer of this problem using a Wirtinger
    calculus based gradient descent with an Armijo-Goldstein step size control,
    which is initialized with the vector ``v`` obtained from the graph
    traversal rank-1 approximation method. (During the iterations it is ensured
    that the least squares error decreases, otherwise the iterations are
    aborted with a warning. Consequently, in terms of the least squares error,
    this approximation is at least as accurate as the traversal-based
    approximation.)

    Parameters
    ----------
    V : .hynet_sparse_
        Matrix that should be approximated by ``vv^H``.
    edges : (numpy.ndarray[.hynet_int_], numpy.ndarray[.hynet_int_])
        Specification of the sparsity pattern of the matrix ``V``.
    roots : numpy.ndarray[.hynet_int_]
        Root nodes for the graph components of the sparsity pattern. The
        absolute angle in ``v`` is adjusted such that the absolute angle
        at the root nodes is zero.
    grad_thres : .hynet_float_, optional
        Absolute threshold on the 2-norm of the objective's gradient w.r.t.
        ``v`` divided by ``N``. (Termination criterion on the first order
        condition for local optimality.)
    mse_rel_thres : .hynet_float_, optional
        Threshold on the relative improvement of ``||P(vv^H - V)||_F^2 / N``
        from the candidate solution ``v`` in iteration ``i`` to ``v`` in
        iteration ``i + 1``. (Termination criterion on stalling progress.)
    max_iter : .hynet_int_, optional
        Maximum number of iterations. (Fallback in case the other termination
        criteria are not met in a reasonable number of iterations.)
    show_convergence_plot : bool, optional
        If True, a plot is shown to illustrate the convergence behavior.

    Returns
    -------
    v : numpy.ndarray
        Vector ``v``, where  ``vv^H`` approximates ``V`` on the sparsity pattern.
    """
    timer = Timer()
    V = V.tocsr()  # Conversion to CSR for indexing
    (e_src, e_dst) = edges
    N = V.shape[0]
    K = len(e_src)
    f = lambda x: calc_rank1approx_mse(V, x, edges) * (N + 2*K)
    gamma = 0.5
    iteration = 0

    v = rank1approx_via_traversal(V, edges, roots)
    mse = calc_rank1approx_mse(V, v, edges)

    if show_convergence_plot:
        iter_mse = [mse]
        iter_grad = [np.nan]

    vv_offd = np.ones(K, dtype=hynet_complex_)
    P_vv = create_sparse_diag_matrix(np.ones(N)) \
         + create_sparse_matrix(e_src, e_dst, vv_offd, N, N) \
         + create_sparse_matrix(e_dst, e_src, vv_offd.conj(), N, N)

    # The case distinction is necessary due to inconsistencies of SciPy...
    P_V = create_sparse_diag_matrix(V.diagonal())
    if K != 0:
        P_V += create_sparse_matrix(e_src, e_dst,
                                    np.asarray(V[e_src, e_dst])[0], N, N) \
             + create_sparse_matrix(e_dst, e_src,
                                    np.asarray(V[e_dst, e_src])[0], N, N)

    while iteration < max_iter:
        np.multiply(v[e_src], v[e_dst].conj(), out=vv_offd)
        P_vv[range(N), range(N)] = np.square(np.abs(v))
        P_vv[e_src, e_dst] = vv_offd
        P_vv[e_dst, e_src] = vv_offd.conj()
        grad_v = 2 * (P_vv - P_V).dot(v)
        grad_v_mse = np.square(np.abs(grad_v)).mean()

        if grad_v_mse <= grad_thres:
            _log.debug("LS rank-1 approx. ~ Threshold on gradient was reached.")
            break

        gamma = get_armijo_step_size(f, v, grad_v, 2 * gamma)
        v_new = v - gamma * grad_v
        mse_new = calc_rank1approx_mse(V, v_new, edges)

        if mse_new > mse:
            _log.warning(("The gradient descent method was aborted after {0} "
                          "iterations and a gradient 'MSE' of {1:.3g} as the "
                          "objective increased.").format(iteration, grad_v_mse))
            break

        if 1 - mse_new/mse < mse_rel_thres:
            _log.debug("LS rank-1 approx. ~ Progress stalled, iteration aborted.")
            break

        if show_convergence_plot:
            iter_mse.append(mse_new)
            iter_grad.append(grad_v_mse)

        (v, mse) = (v_new, mse_new)
        iteration += 1

    if iteration == max_iter:
        _log.debug("LS rank-1 approx. ~ Maximum number of iterations reached.")

    # Adjust absolute phase in all components
    components = get_graph_components(np.arange(N), edges, roots)
    for component in components:
        v[component] *= np.exp(-1j * np.angle(v[component[0]]))

    if show_convergence_plot:
        fig = plt.figure(figsize=(10, 6))
        fig.suptitle("Least-Squares Rank-1 Approximation via Gradient Descent")
        ax = fig.add_subplot(2, 1, 1)
        ax.plot(iter_mse, color='xkcd:sea blue', linestyle='-', linewidth=1)
        ax.margins(x=0)
        ax.set_xlim(left=0)
        ax.set_title("Reconstruction MSE")
        ax = fig.add_subplot(2, 1, 2)
        ax.plot(iter_grad, color='xkcd:sea blue', linestyle='-', linewidth=1)
        ax.margins(x=0)
        ax.set_xlim(left=0)
        ax.set_title("2-norm of the gradient over the number of dimensions")
        ax.set_xlabel('Iteration')
        fig.subplots_adjust(hspace=0.3)
        fig.show()

    _log.debug("LS rank-1 approx. ~ Ended with gradient 'MSE' of "
               "{0:.3e} after {1} iterations ({2:.3f} sec.)"
               .format(grad_v_mse, iteration, timer.total()))
    return v


def get_armijo_step_size(f, x, grad_fx, gamma=1, epsilon=0.2, alpha=2):
    """
    Return a step size based on the Armijo-Goldstein condition.

    This step size is designed for a Wirtinger calculus based gradient descent
    method that minimizes the function ``f(x)``, where ``x in C^N`` is the
    current candidate solution, ``grad_fx`` is the gradient vector w.r.t.
    ``conj(x)`` evaluated at ``x``, and the step direction ``d = -grad_fx``.
    This function returns a step size ``s in R_(+)``, such that the algorithmic
    map reads ``x -> x + s * d``.
    """
    # LITERATURE:
    # 1) Armijo's rule is described, e.g., in "Nonlinear Programming - Theory
    #    and Algorithms" by Bazaraa, Sherali, and Shetty (3rd ed., ch. 8.3).
    # 2) Relevant aspects of Wirtinger calculus are described, e.g., in
    #    "Complex-Valued Adaptive Signal Processing Using Nonlinear Functions"
    #    by Li and Adali (EURASIP Journal on Advances in Signal Processing)

    fx = f(x)
    dd = -epsilon * 2 * np.sum(np.square(np.abs(grad_fx)))

    theta = lambda gamma: f(x - gamma*grad_fx)
    theta_tld = lambda gamma: fx + gamma * dd

    while gamma > 1e-10:
        delta = theta(gamma) - theta_tld(gamma)
        delta_bar = theta(alpha*gamma) - theta_tld(alpha*gamma)
        if delta <= 0 < delta_bar:
            return gamma
        if delta <= 0:
            gamma *= alpha
        else:
            gamma /= alpha

    _log.warning("The step size is very small.")
    return gamma



#pylint: disable=too-many-instance-attributes,too-many-public-methods
"""
Representation of a steady-state scenario in *hynet*.
"""

import logging
import copy
from collections import namedtuple, OrderedDict
import os.path

import numpy as np
import pandas as pd

from hynet.types_ import (hynet_id_,
                          hynet_int_,
                          hynet_float_,
                          hynet_complex_,
                          BusType,
                          BranchType,
                          InjectorType)
from hynet.utilities.base import truncate_with_ellipsis, Timer
from hynet.utilities.graph import (eliminate_parallel_edges,
                                   is_acyclic_graph,
                                   get_graph_components)
from hynet.scenario.capability import CapRegion, ConverterCapRegion
from hynet.scenario.cost import PWLFunction
from hynet.scenario.verification import verify_scenario

_log = logging.getLogger(__name__)


class Scenario:
    """
    Specification of a steady-state grid scenario.

    Parameters
    ----------
    id : .hynet_id_
        Identifier of the scenario.
    name : str
        Name of the scenario.
    time : .hynet_float_
        Time in hours, relative to the scenario collection start time.
    database_uri : str
        URI of the associated *hynet* grid database.
    grid_name : str
        Name of the grid.
    base_mva : .hynet_float_
        Apparent power normalization constant in MVA.
    loss_price : .hynet_float_
        Artificial price for losses in $/MWh. The corresponding cost of losses
        is taken into account for the minimum-cost dispatch.
    description : str
        Description of the grid database.
    annotation : str
        Annotation string for supplementary information.
    bus : pandas.DataFrame
        Data frame with one data set per bus, indexed by the *bus ID*, which
        comprises the following columns:

        ``type``: (``BusType``)
            Type of voltage waveform at the bus.
        ``ref``: (``bool``)
            ``True`` if the bus serves as a reference.
        ``base_kv``: (``hynet_float_``)
            Base voltage in kV.
        ``y_tld``: (``hynet_complex_``)
            Shunt admittance in p.u.. For example, to add a shunt compensator
            that injects ``q`` Mvar and dissipates ``p`` MW at 1 p.u., set
            ``y_tld`` to ``(p + 1j*q) / base_mva``.
        ``load``: (``hynet_complex_``)
            Aggregated inelastic apparent power load in MVA.
        ``v_min``: (``hynet_float_``)
            Voltage lower bound in p.u..
        ``v_max``: (``hynet_float_``)
            Voltage upper bound in p.u..
        ``zone``: (``hynet_id_`` or ``None``)
            Zone ID or ``None`` if not available.
        ``annotation``: (``str``)
            Annotation string for supplementary information.

    branch : pandas.DataFrame
        Data frame with one data set per branch, indexed by the *branch ID*,
        which comprises the following columns:

        ``type``: (``BranchType``)
            Type of entity modeled by the branch.
        ``src``: (``hynet_id_``)
            Source bus ID.
        ``dst``: (``hynet_id_``)
            Destination bus ID.
        ``z_bar``: (``hynet_complex_``)
            Series impedance of the pi-equivalent in p.u..
        ``y_src``: (``hynet_complex_``)
            Shunt admittance at the source side of the pi-equivalent in p.u..
        ``y_dst``: (``hynet_complex_``)
            Shunt admittance at the destination side of the pi-equivalent in
            p.u..
        ``rho_src``: (``hynet_complex_``)
            Complex voltage ratio of the ideal transformer at the source side.
        ``rho_dst``: (``hynet_complex_``)
            Complex voltage ratio of the ideal transformer at the destination
            side.
        ``length``: (``hynet_float_``)
            Line length in kilometers or ``numpy.nan`` if not available.
        ``rating``: (``hynet_float_``)
            Ampacity in terms of a long-term MVA rating at a bus voltage of
            1 p.u. (i.e., the current limit is ``rating/base_mva``)  or
            ``numpy.nan`` if omitted.
        ``angle_min``: (``hynet_float_``)
            Angle difference lower bound in degrees or ``numpy.nan`` if
            omitted.
        ``angle_max``: (``hynet_float_``)
            Angle difference upper bound in degrees or ``numpy.nan`` if
            omitted.
        ``drop_min``: (``hynet_float_``)
            Voltage drop lower bound in percent or ``numpy.nan`` if omitted.
        ``drop_max``: (``hynet_float_``)
            Voltage drop upper bound in percent or ``numpy.nan`` if omitted.
        ``annotation``: (``str``)
            Annotation string for supplementary information.

    converter : pandas.DataFrame
        Data frame with one data set per converter, indexed by the *converter
        ID*, which comprises the following columns:

        ``src``: (``hynet_id_``)
            Source bus ID.
        ``dst``: (``hynet_id_``)
            Destination bus ID.
        ``cap_src``: (``ConverterCapRegion``)
            Specification of the converter's capability region in the P/Q-plane
            (in MW and Mvar, respectively) at the source bus. The P-axis is the
            active power flow *into the converter* and the Q-axis is the
            reactive power injection *into the grid*.
        ``cap_dst``: (``ConverterCapRegion``)
            Specification of the converter's capability region in the P/Q-plane
            (in MW and Mvar, respectively) at the destination bus. The P-axis
            is the active power flow *into the converter* and the Q-axis is the
            reactive power injection *into the grid*.
        ``loss_fwd``: (``hynet_float_``)
            Loss factor in percent for the *forward flow* of active power. This
            proportional loss factor describes the dynamic conversion losses if
            active power flows from the source bus to the destination bus.
        ``loss_bwd``: (``hynet_float_``)
            Loss factor in percent for the *backward flow* of active power. This
            proportional loss factor describes the dynamic conversion losses if
            active power flows from the destination bus to the source bus.
        ``loss_fix``: (``hynet_float_``)
            Static losses in MW (considered at the converter's source bus).
        ``annotation``: (``str``)
            Annotation string for supplementary information.

    injector : pandas.DataFrame
        Data frame with one data set per injector, indexed by the *injector
        ID*, which comprises the following columns:

        ``type``: (``InjectorType``)
            Type of entity modeled by the injector.
        ``bus``: (``hynet_id_``)
            Terminal bus ID.
        ``cap``: (``CapRegion``)
            Specification of the injector's capability region in the P/Q-plane
            (in MW and Mvar, respectively).
        ``cost_p``: (``PWLFunction`` or ``None``)
            Piecewise linear cost function for active power, which specifies
            the cost in dollars for an active power injection in MW, or
            ``None`` in case of zero costs.
        ``cost_q``: (``PWLFunction`` or ``None``)
            Piecewise linear cost function for reactive power, which specifies
            the cost in dollars for a reactive power injection in Mvar, or
            ``None`` in case of zero costs.
        ``cost_start``: (``hynet_float_``)
            Startup cost in dollars.
        ``cost_stop``: (``hynet_float_``)
            Shutdown cost in dollars.
        ``ramp_up``: (``hynet_float_``)
            Maximum upramping rate for active power in MW/h or ``numpy.nan`` if
            the upramping rate is not limited.
        ``ramp_down``: (``hynet_float_``)
            Maximum downramping rate for active power in MW/h or ``numpy.nan``
            if the downramping rate is not limited.
        ``min_up``: (``hynet_float_``)
            Minimum uptime in hours or ``numpy.nan`` if it is not limited.
        ``min_down``: (``hynet_float_``)
            Minimum downtime in hours or ``numpy.nan`` if it is not limited.
        ``energy_min``: (``hynet_float_``)
            Energy lower bound in MWh or ``numpy.nan`` if it is not limited.
        ``energy_max``: (``hynet_float_``)
            Energy upper bound in MWh or ``numpy.nan`` if it is not limited.
        ``annotation``: (``str``)
            Annotation string for supplementary information.

    See Also
    --------
    hynet.data.interface.load_scenario:
        Load a scenario from a *hynet* grid database.
    """
    def __init__(self):
        self.id = 0
        self.name = ''
        self.time = 0
        self.database_uri = ''
        self.grid_name = ''
        self.base_mva = np.nan
        self.loss_price = 0
        self.description = ''
        self.annotation = ''

        # The data frames are initialized with empty arrays to indicate the
        # data type and, in case of the absence of converters, this is also
        # necessary to avoid warnings from SciPy. To preserve the column
        # order, the initialization is performed via an ordered dictionary.
        # (This necessitates pandas 0.23.0 or higher.)

        self.bus = pd.DataFrame(OrderedDict(
            [('type', np.ndarray(0, dtype=BusType)),
             ('ref', np.ndarray(0, dtype=bool)),
             ('base_kv', np.ndarray(0, dtype=hynet_float_)),
             ('y_tld', np.ndarray(0, dtype=hynet_complex_)),
             ('load', np.ndarray(0, dtype=hynet_complex_)),
             ('v_min', np.ndarray(0, dtype=hynet_float_)),
             ('v_max', np.ndarray(0, dtype=hynet_float_)),
             ('zone', np.ndarray(0, dtype=hynet_id_)),
             ('annotation', np.ndarray(0, dtype=str))
             ]))
        self.bus.index = pd.Index([], name='id', dtype=hynet_id_)

        self.branch = pd.DataFrame(OrderedDict(
            [('type', np.ndarray(0, dtype=BranchType)),
             ('src', np.ndarray(0, dtype=hynet_id_)),
             ('dst', np.ndarray(0, dtype=hynet_id_)),
             ('z_bar', np.ndarray(0, dtype=hynet_complex_)),
             ('y_src', np.ndarray(0, dtype=hynet_complex_)),
             ('y_dst', np.ndarray(0, dtype=hynet_complex_)),
             ('rho_src', np.ndarray(0, dtype=hynet_complex_)),
             ('rho_dst', np.ndarray(0, dtype=hynet_complex_)),
             ('length', np.ndarray(0, dtype=hynet_float_)),
             ('rating', np.ndarray(0, dtype=hynet_float_)),
             ('angle_min', np.ndarray(0, dtype=hynet_float_)),
             ('angle_max', np.ndarray(0, dtype=hynet_float_)),
             ('drop_min', np.ndarray(0, dtype=hynet_float_)),
             ('drop_max', np.ndarray(0, dtype=hynet_float_)),
             ('annotation', np.ndarray(0, dtype=str))
             ]))
        self.branch.index = pd.Index([], name='id', dtype=hynet_id_)

        self.converter = pd.DataFrame(OrderedDict(
            [('src', np.ndarray(0, dtype=hynet_id_)),
             ('dst', np.ndarray(0, dtype=hynet_id_)),
             ('cap_src', np.ndarray(0, dtype=ConverterCapRegion)),
             ('cap_dst', np.ndarray(0, dtype=ConverterCapRegion)),
             ('loss_fwd', np.ndarray(0, dtype=hynet_float_)),
             ('loss_bwd', np.ndarray(0, dtype=hynet_float_)),
             ('loss_fix', np.ndarray(0, dtype=hynet_float_)),
             ('annotation', np.ndarray(0, dtype=str))
             ]))
        self.converter.index = pd.Index([], name='id', dtype=hynet_id_)

        self.injector = pd.DataFrame(OrderedDict(
            [('type', np.ndarray(0, dtype=InjectorType)),
             ('bus', np.ndarray(0, dtype=hynet_id_)),
             ('cap', np.ndarray(0, dtype=CapRegion)),
             ('cost_p', np.ndarray(0, dtype=PWLFunction)),
             ('cost_q', np.ndarray(0, dtype=PWLFunction)),
             ('cost_start', np.ndarray(0, dtype=hynet_float_)),
             ('cost_stop', np.ndarray(0, dtype=hynet_float_)),
             ('ramp_up', np.ndarray(0, dtype=hynet_float_)),
             ('ramp_down', np.ndarray(0, dtype=hynet_float_)),
             ('min_up', np.ndarray(0, dtype=hynet_float_)),
             ('min_down', np.ndarray(0, dtype=hynet_float_)),
             ('energy_min', np.ndarray(0, dtype=hynet_float_)),
             ('energy_max', np.ndarray(0, dtype=hynet_float_)),
             ('annotation', np.ndarray(0, dtype=str))
             ]))
        self.injector.index = pd.Index([], name='id', dtype=hynet_id_)

    def __eq__(self, other):
        """Return True if the scenarios are equal."""
        def fix_sorting(dataframe):
            return dataframe.sort_index(axis='columns').sort_index(axis='index')

        def compare_dataframe(dataframe1, dataframe2):
            return fix_sorting(dataframe1).equals(fix_sorting(dataframe2))

        # REMARK: The complex voltage ratios need to be treated separately
        # as they are converted between polar and rectangular form during
        # loading and saving, which causes slight numerical differences.
        rho_columns = ['rho_src', 'rho_dst']
        if self.branch.shape == other.branch.shape:
            rho_equal = \
                np.all(np.isclose(fix_sorting(self.branch[rho_columns]),
                                  fix_sorting(other.branch[rho_columns])))
        else:
            rho_equal = False
        self_branch = self.branch.drop(rho_columns, axis='columns')
        other_branch = other.branch.drop(rho_columns, axis='columns')

        return (rho_equal and
                self.id == other.id and
                self.name == other.name and
                self.time == other.time and
                self.database_uri == other.database_uri and
                self.grid_name == other.grid_name and
                self.base_mva == other.base_mva and
                self.loss_price == other.loss_price and
                self.description == other.description and
                self.annotation == other.annotation and
                compare_dataframe(self.bus, other.bus) and
                compare_dataframe(self_branch, other_branch) and
                compare_dataframe(self.converter, other.converter) and
                compare_dataframe(self.injector, other.injector))

    def copy(self):
        """Return a deep copy of this scenario."""
        timer = Timer()
        scr = copy.deepcopy(self)
        # The deep copy on object columns is not supported, see
        #   https://github.com/pandas-dev/pandas/issues/12663
        # Due to this, we deep copy those manually.
        for l, (cap_src, cap_dst) in enumerate(zip(self.converter['cap_src'],
                                                   self.converter['cap_dst'])):
            id_ = self.converter.index[l]
            scr.converter.at[id_, 'cap_src'] = copy.deepcopy(cap_src)
            scr.converter.at[id_, 'cap_dst'] = copy.deepcopy(cap_dst)
        for j, (cap, cost_p, cost_q) in enumerate(zip(self.injector['cap'],
                                                      self.injector['cost_p'],
                                                      self.injector['cost_q'])):
            id_ = self.injector.index[j]
            scr.injector.at[id_, 'cap'] = copy.deepcopy(cap)
            scr.injector.at[id_, 'cost_p'] = copy.deepcopy(cost_p)
            scr.injector.at[id_, 'cost_q'] = copy.deepcopy(cost_q)
        _log.debug("Scenario deep copy ({:.3f} sec.)".format(timer.total()))
        return scr

    def verify(self):
        """
        Verify the integrity and validity of the scenario.

        This method performs an extensive series of checks on the scenario
        to ensure that the data is consistent (e.g., the references between
        data frames), proper (e.g., constraint limits), and valid (i.e.,
        compliant with all preconditions as assumed by *hynet*).

        Besides errors being reported via an exception, critical settings are
        reported via a warning and adverse settings w.r.t. to the exactness of
        the relaxations are issued via an info on the log. To enable these
        messages, the logging system can be configured as

        >>> import logging
        >>> logging.basicConfig(level=logging.INFO)

        Raises
        ------
        ValueError
            In case any kind of integrity or validity violation is detected.
        """
        verify_scenario(self)

    @property
    def num_buses(self):
        """Return the number of buses."""
        return len(self.bus.index)

    @property
    def num_branches(self):
        """Return the number of branches."""
        return len(self.branch.index)

    @property
    def num_converters(self):
        """Return the number of converters."""
        return len(self.converter.index)

    @property
    def num_injectors(self):
        """Return the number of injectors."""
        return len(self.injector.index)

    @property
    def e_src(self):
        """
        Return the branch source bus indices as a pandas series.

        **Remark:** This property is *index-based* and intended for internal use.
        """
        return pd.Series(self.get_bus_index(self.branch['src'].values),
                         index=self.branch.index)

    @property
    def e_dst(self):
        """
        Return the branch destination bus indices as a pandas series.

        **Remark:** This property is *index-based* and intended for internal use.
        """
        return pd.Series(self.get_bus_index(self.branch['dst'].values),
                         index=self.branch.index)

    @property
    def c_src(self):
        """
        Return the converter source bus indices as a pandas series.

        **Remark:** This property is *index-based* and intended for internal use.
        """
        return pd.Series(self.get_bus_index(self.converter['src'].values),
                         index=self.converter.index)

    @property
    def c_dst(self):
        """
        Return the converter destination bus indices as a pandas series.

        **Remark:** This property is *index-based* and intended for internal use.
        """
        return pd.Series(self.get_bus_index(self.converter['dst'].values),
                         index=self.converter.index)

    @property
    def n_src(self):
        """
        Return the injector terminal bus indices as a pandas series.

        **Remark:** This property is *index-based* and intended for internal use.
        """
        return pd.Series(self.get_bus_index(self.injector['bus'].values),
                         index=self.injector.index)

    def get_bus_index(self, bus_id):
        """
        Return the bus index for the given bus identifier.

        **Remark:** This result is *index-based* and intended for internal use.
        """
        return np.array([self.bus.index.get_loc(key) for key in bus_id],
                        dtype=hynet_int_)

    def get_ref_buses(self):
        """
        Return an array with the bus index of the reference buses.

        **Remark:** This result is *index-based* and intended for internal use.
        """
        return np.nonzero(self.bus['ref'].values)[0]

    def has_acyclic_subgrids(self):
        """
        Return ``True`` if all subgrids are acyclic and ``False`` otherwise.
        """
        return is_acyclic_graph(np.arange(self.num_buses),
                                (self.e_src.values, self.e_dst.values))

    def has_acyclic_ac_subgrids(self):
        """
        Return ``True`` if all AC subgrids are acyclic and ``False`` otherwise.
        """
        ac_buses = np.where(np.equal(self.bus['type'].values, BusType.AC))[0]
        mask = np.logical_and(np.isin(self.e_src.values, ac_buses),
                              np.isin(self.e_dst.values, ac_buses))
        return is_acyclic_graph(ac_buses, (self.e_src.values[mask],
                                           self.e_dst.values[mask]))

    def has_acyclic_dc_subgrids(self):
        """
        Return ``True`` if all DC subgrids are acyclic and ``False`` otherwise.
        """
        dc_buses = np.where(np.equal(self.bus['type'].values, BusType.DC))[0]
        mask = np.logical_and(np.isin(self.e_src.values, dc_buses),
                              np.isin(self.e_dst.values, dc_buses))
        return is_acyclic_graph(dc_buses, (self.e_src.values[mask],
                                           self.e_dst.values[mask]))

    def get_ac_subgrids(self):
        """
        Return a list with a pandas index of bus IDs for every AC subgrid.
        """
        ac_buses = np.where(np.equal(self.bus['type'].values, BusType.AC))[0]
        mask = np.logical_and(np.isin(self.e_src.values, ac_buses),
                              np.isin(self.e_dst.values, ac_buses))
        subgrids = get_graph_components(ac_buses, (self.e_src.values[mask],
                                                   self.e_dst.values[mask]))
        return [self.bus.index[subgrid] for subgrid in subgrids]

    def get_dc_subgrids(self):
        """
        Return a list with a pandas index of bus IDs for every DC subgrid.
        """
        dc_buses = np.where(np.equal(self.bus['type'].values, BusType.DC))[0]
        mask = np.logical_and(np.isin(self.e_src.values, dc_buses),
                              np.isin(self.e_dst.values, dc_buses))
        subgrids = get_graph_components(dc_buses, (self.e_src.values[mask],
                                                   self.e_dst.values[mask]))
        return [self.bus.index[subgrid] for subgrid in subgrids]

    def get_islands(self):
        """
        Return a list with a pandas index of bus IDs for every islanded grid.
        """
        islands = get_graph_components(np.arange(self.num_buses),
                                       (np.concatenate((self.e_src.values,
                                                        self.c_src.values)),
                                        np.concatenate((self.e_dst.values,
                                                        self.c_dst.values))))
        return [self.bus.index[island] for island in islands]

    def analyze_cycles(self):
        """
        Return a data frame with an analysis of cyclic connections of buses.

        The information about cyclic connections of buses is relevant in the
        study of transitions to the *hybrid architecture*, which establishes
        exactness of the semidefinite and second-order cone relaxation of the
        OPF problem under normal operating conditions, cf. [1]_.

        Returns
        -------
        result : pandas.DataFrame
            Data frame with the cycle analysis result for every subgrid, which
            comprises the following columns:

            ``num_cycles``: (``hynet_int_``)
                Number of cycles in the subgrid.
            ``type``: (``BusType``)
                Type of the subgrid.
            ``buses``: (``pandas.Index``)
                Pandas index with the *bus IDs* of the buses that are part of
                the subgrid.

        References
        ----------
        .. [1] M. Hotz and W. Utschick, "The Hybrid Transmission Grid
               Architecture: Benefits in Nodal Pricing," in IEEE Trans. Power
               Systems, vol. 33, no. 2, pp. 1431-1442, Mar. 2018.
        """
        subgrids = self.get_ac_subgrids() + self.get_dc_subgrids()
        src = self.branch['src']
        dst = self.branch['dst']

        result = pd.DataFrame(OrderedDict(
            [('num_cycles', np.ndarray(0, dtype=hynet_int_)),
             ('type', np.ndarray(0, dtype=BusType)),
             ('buses', np.ndarray(0, dtype=object))
             ]))
        result.index = pd.Index([], name='subgrid', dtype=hynet_id_)

        for n, subgrid in enumerate(subgrids):
            idx = src.isin(subgrid) | dst.isin(subgrid)
            num_corridors = len(eliminate_parallel_edges((src.loc[idx].values,
                                                          dst.loc[idx].values))[0])
            result.loc[n + 1] = (num_corridors - (len(subgrid) - 1),
                                 self.bus.at[subgrid[0], 'type'],
                                 subgrid)
        return result

    def get_branches_in_corridors(self, corridors):
        """
        Return a pandas index of branch IDs that reside in these corridors.

        **Remark:** This property is *index-based* and intended for internal use.

        Parameters
        ----------
        corridors : (numpy.ndarray[.hynet_int_], numpy.ndarray[.hynet_int_])
            Tuple of NumPy arrays that state the source *bus index* and
            destination *bus index* of the corridors.

        Returns
        -------
        index : pandas.Index
            Pandas index with the *branch IDs* of those branches that reside in
            the specified corridors.
        """
        if self.branch.empty or not corridors[0].size:
            return pd.Index([], dtype=hynet_id_, name=self.branch.index.name)

        edges = (self.e_src.values, self.e_dst.values)
        E = np.column_stack((edges[0], edges[1]))
        E.sort(axis=1)

        C = np.column_stack((corridors[0], corridors[1]))
        C.sort(axis=1)

        resident = np.zeros(self.num_branches, dtype=bool)
        for n in range(C.shape[0]):
            resident |= (E[:, 0] == C[n, 0]) & (E[:, 1] == C[n, 1])
        return self.branch.index[resident]

    def _get_parallel_branch_indices(self):
        """
        Return a list with an array of *branch indices* for parallel branches.
        """
        parallel_branches = []
        if self.branch.empty:
            return parallel_branches

        E = np.column_stack((self.e_src.values, self.e_dst.values))
        E.sort(axis=1)

        (C, count) = np.unique(E, axis=0, return_counts=True)
        C = C[count > 1]
        for n in range(C.shape[0]):
            parallel_branches.append(
                np.where((E[:, 0] == C[n, 0]) & (E[:, 1] == C[n, 1]))[0])
        return parallel_branches

    def get_parallel_branches(self):
        """
        Return a list with a pandas index of branch IDs for parallel branches.
        """
        index = self.branch.index
        return [index[x] for x in self._get_parallel_branch_indices()]

    def remove_buses(self, buses):
        """
        Remove the specified buses and attached branches, converters, and injectors.

        Parameters
        ----------
        buses : Iterable[.hynet_id_]
            Iterable of bus IDs that specifies the buses to be removed.

        Returns
        -------
        branches : pandas.Index
            Removed branches, which were connected to the removed buses.
        converters : pandas.Index
            Removed converters, which were connected to the removed buses.
        injectors : pandas.Index
            Removed injectors, which were connected to the removed buses.
        """
        bus_idx = pd.Index(buses)
        self.bus.drop(bus_idx, axis=0, inplace=True)

        bus_lst = bus_idx.tolist()

        branches = self.branch.index[self.branch['src'].isin(bus_lst) |
                                     self.branch['dst'].isin(bus_lst)]
        self.branch.drop(branches, axis=0, inplace=True)

        converters = self.converter.index[self.converter['src'].isin(bus_lst) |
                                          self.converter['dst'].isin(bus_lst)]
        self.converter.drop(converters, axis=0, inplace=True)

        injectors = self.injector.index[self.injector['bus'].isin(bus_lst)]
        self.injector.drop(injectors, axis=0, inplace=True)
        return branches, converters, injectors

    def get_time_tuple(self):
        """
        Return the scenario time stamp as ``(days, hours, minutes, seconds)``.
        """
        (min_, sec_) = divmod(round(self.time * 3600), 60)
        (hrs_, min_) = divmod(min_, 60)
        (dys_, hrs_) = divmod(hrs_, 24)
        time_ = namedtuple('time', ['days', 'hours', 'minutes', 'seconds'])
        return time_(hynet_int_(dys_), hynet_int_(hrs_),
                     hynet_int_(min_), hynet_int_(sec_))

    def get_time_string(self):
        """
        Return the scenario time stamp as ``Dd HH:MM:SS``.
        """
        return "{0.days}d {0.hours:02d}:{0.minutes:02d}:{0.seconds:02d}"\
               .format(self.get_time_tuple())

    def set_conservative_rating(self):
        """
        Adjust the branch rating to a conservative setting (MATPOWER compat.).

        In the *hynet* data format, the branch rating is an ampacity rating in
        terms of the apparent power flow at 1 p.u. (due to reasons stated in
        [1]_, Section III), i.e., the rating divided by ``base_mva`` is the
        ampacity rating. In the MATPOWER format, the rating is an apparent
        power rating, i.e., it is independent of the voltage. To ensure
        feasibility w.r.t. MATPOWER's apparent power rating, the *hynet* rating
        may be converted to a more conservative rating as described in [1]_,
        Remark 1, which is performed by this method.

        References
        ----------
        .. [1] M. Hotz and W. Utschick, "A Hybrid Transmission Grid
               Architecture Enabling Efficient Optimal Power Flow," in IEEE
               Trans. Power Systems, vol. 31, no. 6, pp. 4504-4516, Nov. 2016.
        """
        v_max = self.bus['v_max']
        v_max_branch = np.maximum(v_max.loc[self.branch['src']].values,
                                  v_max.loc[self.branch['dst']].values)
        self.branch['rating'] /= v_max_branch

    def set_minimum_series_resistance(self, min_resistance):
        """
        Set the series resistance of all branches to a minimum of ``min_resistance``.

        Very small values of the series resistance of branches may lead to
        numerical issues during the solution of the OPF problem. With this
        method, the series resistance is enforced to be ``>= min_resistance``.

        Parameters
        ----------
        min_resistance : float
            Minimum resistance for the series resistance of all branches. If a
            branch has a series resistance below this value, it is replaced by
            ``min_resistance``.
        """
        idx = self.branch['z_bar'].real < min_resistance
        self.branch.loc[idx, 'z_bar'] =\
            min_resistance + 1j*self.branch.loc[idx, 'z_bar'].imag

    def ensure_reference(self):
        """
        Automatically add a reference bus to all AC subgrids without a reference.

        All AC subgrids are required to have a dedicated reference bus.
        However, cases may arise in which certain AC subgrids have no
        predefined reference, e.g., due to the partitioning or islanding of AC
        subgrids in a simulated branch outage. This method automatically
        assigns a reference bus to all AC subgrids without a reference, where
        the reference bus is set to the bus with the injector of highest
        capacity therein, if available.
        """
        for subgrid in self.get_ac_subgrids():
            if self.bus.loc[subgrid, 'ref'].any():
                continue
            injectors = self.injector.loc[self.injector['bus'].isin(subgrid)]
            ref_bus = subgrid[0]
            p_max = 0
            for id_, (bus, cap) in injectors[['bus', 'cap']].iterrows():
                if cap.p_max > p_max:
                    ref_bus = bus
                    p_max = cap.p_max
            self.bus.at[ref_bus, 'ref'] = True

    def __repr__(self):
        return ("\n+" + "-"*78 + "+\n"
                "| SCENARIO" + " "*69 + "|\n"
                "+" + "-"*78 + "+\n"
                "| Grid name:      {1:<61s}" + "|\n"
                "| Database:       {2:<61s}" + "|\n"
                "| Scenario ID:    {3:<61s}" + "|\n"
                "| Scenario name:  {4:<61s}" + "|\n"
                "| Scenario time:  {5:<61s}" + "|\n"
                "| MVA base:       {6:<61s}" + "|\n"
                "| Loss price:     {7:<61s}" + "|\n"
                "+" + "-"*78 + "+\n\n"
                "+" + "-"*78 + "+\n"
                "| BUS DATA" + " "*69 + "|\n"
                "+" + "-"*78 + "+\n\n{0.bus}\n\n"
                "+" + "-"*78 + "+\n"
                "| BRANCH DATA" + " "*66 + "|\n"
                "+" + "-"*78 + "+\n\n{0.branch}\n\n"
                "+" + "-"*78 + "+\n"
                "| CONVERTER DATA" + " "*63 + "|\n"
                "+" + "-"*78 + "+\n\n{0.converter}\n\n"
                "+" + "-"*78 + "+\n"
                "| INJECTOR DATA" + " "*64 + "|\n"
                "+" + "-"*78 + "+\n\n{0.injector}\n"
                ).format(self,
                         truncate_with_ellipsis(self.grid_name, 61),
                         truncate_with_ellipsis(
                             os.path.basename(self.database_uri), 61),
                         str(self.id),
                         truncate_with_ellipsis(self.name, 61),
                         self.get_time_string(),
                         str(self.base_mva) + " MVA",
                         str(self.loss_price) + " $/MWh")



"""
Regression test to verify the proper installation of *hynet* at the user.
"""

from hynet.types_ import SolverType, SolverStatus
from hynet.solver import AVAILABLE_SOLVERS
from hynet.opf.calc import calc_opf
from hynet.test.regression import OPFVerificationData, verify_opf_result
from hynet.test.system import TEST_SYSTEM, REFERENCE_SOLUTION


def test_installation(verbose=True):  # pylint: disable=too-many-branches
    """
    Verify the OPF solution of a test system for all available solvers.

    Parameters
    ----------
    verbose : bool, optional
        If True (default), the test is documented to the standard output.

    Returns
    -------
    success : bool
        True if the test was passed, False otherwise.
    """
    if not AVAILABLE_SOLVERS:
        if verbose:
            print("ERROR: No supported solvers were found.")
        return False

    try:
        solvers = [solver_class() for solver_class in AVAILABLE_SOLVERS]
    except Exception as exception:  # pylint: disable=broad-except
        if verbose:
            print("ERROR: " + str(exception))
        return False

    success = True
    width = max([len(str(solver)) for solver in solvers]) + 3
    for solver in solvers:
        solver_name = str(solver)
        message = solver_name + ' ' + '.'*(width - len(solver_name)) + ' '

        try:
            result = calc_opf(TEST_SYSTEM, solver=solver)
        except Exception as exception:  # pylint: disable=broad-except
            if verbose:
                print(message + "ERROR: " + str(exception))
            success = False
            continue

        if result.solver_status != SolverStatus.SOLVED:
            message += ("ERROR: Solver failed with status '" +
                        result.solver_status.name + "'.")
            success = False
        elif result.empty:
            message += "ERROR: The solver did not return any result data."
            success = False
        else:
            tolerance = 5e-4 if solver.type == SolverType.QCQP else 5e-3
            try:
                verify_opf_result(OPFVerificationData.create(result),
                                  REFERENCE_SOLUTION,
                                  tolerance)
            except ValueError as exception:
                message += "ERROR: " + str(exception)
                success = False
            else:
                message += "OK"
        if verbose:
            print(message)

    return success
